/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <assert.h>

/* LibEtPan! provides this */
#include <clist.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"
#include "defines.h"
#include "conf.h"
#include "cntree.h"
#include "lxml.h"
#include "utils.h"

/* Static global variable for XePConf, as there is only one 
 * instance of XePConf at a time, this is safe.
 */
/*
 * TODO: clean up so we do not need any pointer to conf other than
 * this one.
 */
static XePConf *global_conf;

static XePConf *
Conf_new(void);

static cntree *
Conf_parse_rc_file(const char *path, const char *rcfile);

static int
Conf_write_rc_file(const char *path,
		   const char *rcfile, 
		   const char *dtd, 
		   cntree *ntree);

#if 0
static XePPreferences *
Conf_read_preferences(const char *path);
#endif

static XePConfiguration *
Conf_read_configuration(const char *path);

static int
Conf_write_configuration(const char *path, XePConfiguration *configuration);

/* Function definitions */
static XePConf *
Conf_new(void)
{
  XePConf *conf;

  conf = x_alloc(1, XePConf);
  if (conf == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      conf->preferences = NULL;
      conf->configuration = NULL;
      conf->folderslist = NULL;
    }

  global_conf = conf;
  return conf;
}

void
a_Conf_free(XePConf *conf)
{
  if (!conf)
    return;
  A_LOGGING_DEBUG("freeing config: %p", conf);
  
  a_Configuration_free(conf->configuration);

  if (conf->path)
    free(conf->path);
  free(conf);

  return;
}

/*! Return a the full path of the configuration file "file"
 *  If "path" is NULL, get the configuration path from the
 *  environment variables XETPAN_CONFIG, or from $HOME"/xetpan".
 *  If none of XETPAN_CONFIG and HOME are set, look into current directory.
 */
char *
a_Conf_create_filename(const char *path, const char *file)
{
  char *rcpath;
  char *configrc;

  assert(file != NULL);
  
  if (path == NULL) 
    {
      char *tmp;
      tmp = getenv("XETPAN_CONFIG");
      if (tmp)
	rcpath = strdup(tmp);
      else
	{
	  tmp = getenv("HOME");
	  if (!tmp)
	    tmp = ".";
	  rcpath = (char *) malloc(strlen(tmp) + 
				   strlen(XETPAN_DIR_SEPARATOR_S) +
				   strlen(XETPAN_CONF_DIR) + 1);
	  sprintf(rcpath, "%s%s%s", tmp, XETPAN_DIR_SEPARATOR_S,
		  XETPAN_CONF_DIR);
	}
    }
  else
    rcpath = strdup(path);
    
  configrc = (char *) malloc(strlen(rcpath) + 
			     strlen(XETPAN_DIR_SEPARATOR_S) + 
			     strlen(file) + 1);
  sprintf(configrc, "%s%s%s", rcpath, XETPAN_DIR_SEPARATOR_S, file);

  free(rcpath);

  return configrc;
}

XePConf *
a_Conf_read(const char *path)
{
  XePConfiguration *configuration = NULL;
  XePConf          *conf          = NULL;

  if ((configuration = Conf_read_configuration(path)) != NULL)
    {
      conf = Conf_new();
      if (conf == NULL)
	{
	  a_Configuration_free(configuration);
	}
      else
	{
	  conf->configuration = configuration;
	  
	  if (path == NULL)
	    conf->path = a_Conf_create_filename(path, "");
	  else
	    conf->path = strdup(path);
	  if (conf->path == NULL)
	    {
	      A_LOGGING_ERROR("malloc: %s", strerror(errno));
	      a_Conf_free(conf);
	      conf = NULL;
	    }
	}
    }
  global_conf = conf;
  return conf;
}

int
a_Conf_write(void)
{
  int r;
  A_LOGGING_TRACE;

  if (global_conf == NULL || global_conf->configuration == NULL)
    {
      A_LOGGING_ERROR("Attempt to write an empty configuration. Configuration not saved.");
      return -1;
    }

  r = Conf_write_configuration(global_conf->path, global_conf->configuration);

  return r;
}

/*! Parse a resources file.
 * \param path where to search. NULL is default config path.
 * \param rcfile basename of the file to parse
 * \return a cntree corresponding to the XML hiearchy.
 */
static cntree *
Conf_parse_rc_file(const char *path, const char *rcfile)
{
  LXMLhandle *handle;
  cntree *config = NULL;
  char *rcpath;

  rcpath = a_Conf_create_filename(path, rcfile);

  A_LOGGING_INFO("Parsing file %s", rcpath);

  handle = lxml_new();
  if(!(lxml_set_filename(handle, rcpath) || lxml_file_open(handle, LXML_Read)))
    {
      config = lxml_file_read(handle);
      lxml_file_close(handle);
    }
  lxml_free(handle);
  free(rcpath);
  return config;
}

static int
Conf_write_rc_file(const char *path,
		   const char *rcfile, 
		   const char *dtd, 
		   cntree *ntree)
{
  LXMLhandle *handle;
  char *rcpath;
  char *newpath;
  const char *new_ext = ".new";
  int r = -1;

  /* Write .new file, and if successful, rename old .xml file to .bak 
   * and rename .new to .xml
   */

  if (mkdir(path, S_IRUSR | S_IWUSR | S_IXUSR))
    {
      if (errno != EEXIST)
	{
	  A_LOGGING_ERROR(_("Could not create configuration directory %s: %s"),
			  path, strerror(errno));
	  return -1;
	}
    }

  rcpath = a_Conf_create_filename(path, rcfile);

  newpath = x_alloc(strlen(rcpath) + strlen(new_ext) + 1, char);

  strcpy(newpath, rcpath);
  strcat(newpath, new_ext);

  handle = lxml_new();
  lxml_set_encoding(handle, "UTF-8");
  lxml_set_dtd(handle, dtd);
  if (!(lxml_set_filename(handle, newpath) || lxml_file_open(handle, LXML_Write)))
    {
      r = lxml_file_write(handle, ntree);
      if (r)
	{
	  A_LOGGING_ERROR("Could not successfully write '%s'", rcpath);
	}
      lxml_file_close(handle);
    }
  else
    {
      A_LOGGING_ERROR("Could not open file '%s'. New configuration not written.", newpath);
    }
  lxml_free(handle);

  if (r == 0)
    {
      r = a_Utils_file_backup(rcpath);
      if (r == 0)
	{
	  r = rename(newpath, rcpath);
	  if (r)
	    {
	      A_LOGGING_ERROR("Could not rename '%s' to '%s'. The new "
			      "configuration file is still in '%s'. You "
			      "should rename it to '%s'", newpath, rcpath,
			      newpath, rcpath);
	    }
	}
      else
	{
	  A_LOGGING_ERROR("Could not make backup of current configuration."
			  "New configuration is still in '%s'. To use new "
			  "configuration, you should rename it to '%s'.",
			  newpath, rcpath);
	}
    }

  free(newpath);
  free(rcpath);
  return r;      
}

#if 0
/*! Parse preferences.xep 
 *  \param path where to search
 *  \return preferences structure
 */
static XePPreferences *
Conf_read_preferences(const char *path)
{
  XePPreferences *preferences = NULL;
  cntree *ntree;
  
  ntree = Conf_parse_rc_file(path, XETPAN_PREFERENCES_RC);

  if (ntree)
    preferences = a_Preferences_create_from_tree(ntree);

  lxml_tree_free(ntree);

  return preferences;
}
#endif

/*! Parse configuration.xep 
 *  \param path where to search
 *  \return configuration structure
 */
static XePConfiguration *
Conf_read_configuration(const char *path)
{
  XePConfiguration *configuration = NULL;
  cntree *ntree;
  
  ntree = Conf_parse_rc_file(path, XETPAN_CONFIGURATION_RC);

  configuration = a_Configuration_create_from_tree(ntree);

  lxml_tree_free(ntree);
  return configuration;
}

static int
Conf_write_configuration(const char *path, XePConfiguration *configuration)
{
  int r;

  cntree *ntree;

  ntree = a_Configuration_create_tree(configuration);
  if (ntree)
    {
      r = Conf_write_rc_file(path, XETPAN_CONFIGURATION_RC, 
			     XETPAN_CONFIGURATION_DTD, ntree);
      lxml_tree_free(ntree);
      if (r)
	{
	  A_LOGGING_ERROR("Configuration file not written.");
	}
    }
  else
    {
      A_LOGGING_ERROR("Could not create configuration tree. Configuration not saved.");
      r = -1;
    }

  return r;
}

