/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef FOLDER_H
#define FOLDER_H

#include <clist.h>
#include <chash.h>

#include "cntree.h"
#include "libetpan.h"

#include "manager.h"

#include "folder-types.h"
#include "mailbox-types.h"

typedef clist XePFoldersList;

typedef void (*XePFolderAppendMessagesCb)(carray *folders,
					  carray *messages, 
					  uint32_t failures,
					  void *user_data);

typedef void (*XePFolderListCopyCb)(void *user_data);

typedef struct XePFolderProcessing_s XePFolderProcessing;

/*! Type of folder. */
typedef enum 
  {
    XEP_FOLDER_TYPE_REAL,
    XEP_FOLDER_TYPE_VIRTUAL,
    XEP_FOLDER_TYPE_MAILBOX
  } XePFolderType;

/*! A folder.
 *  This structure holds the messages.
 *  Each folder is managed by a different thread.
 *  A folder can be of one of the following types:
 *     - R-Folder, or real folder. It contains messages;
 *     - V-Folder, or virtual folder.  It contains dynamic search results;
 *     - M-Folder, or mailbox folder.  Just materializes a mailbox.
 */

struct XePFolder_s
{
  /* First member must be id, second must be path.
   * This is so the XePGtkUtilsTree can use a generic structure
   * wich two members are the first and second column of the tree
   */
  char *id;
  char *path;          /**< Path in the mailbox hieararchy, relative
		        *   to its toplever directory. Only for MH,
		        *   NNTP, IMAP and MBOX hierarchies.      */
  XePFolderType type;
  long uid;            /**< Unique Identifier                     */
  XePManagerThread thread;

  XePMailbox *mailbox; /**< Mailbox to which it is related.       */
  clist *rule_refs;    /**< List of processing rules              */

  struct mail_storage *ep_storage;     /**< The storage           */
  struct mail_folder  *ep_folder;      /**< The message folder    */
  struct mail_envelopes_tree *ep_tree; /**< The message tree      */
  struct mail_envelopes_list *ep_list; /**< The message list      */

  clist *folder_refs;    /**< VFolder: list of folders to seach   */
  clist *condition_refs; /**< VFolder: list of conditions         */
  uint32_t total;
  uint32_t unseen;
  uint32_t recent;
};

clist *
a_Folder_list_new(void);

clist *
a_Folder_list_new_copy(clist *folders);

/*! Releases resources from a folder list.
 *  \param folders the list of folders to be deleted.
 */
void
a_Folder_list_free(clist *folders);

void
a_Folder_list_free_copy(clist *folders);

/*! Copy the list of folders from source to target
 *  Disconnect and free source folders marked for delete (uid < 0)
 *  Register new folders
 *  Call user_cb when all changes are done
 *  \param source the list of folders to copy
 *  \param target the list that will receive the copies
 *  \param mailbox the target mailbox
 *  \param user_cb a user function called when operation finishes
 *  \param user_data the user data to pass to user_cb() as second arg
 */
void
a_Folder_list_copy(clist *source, 
		   clist *target,
		   XePMailbox *mailbox,
		   XePFolderListCopyCb user_cb,
		   void *user_data);

/*! Create a new folder and register it 
 *  The newly created folder has:
 *  - a unique strictly positive uid
 *  - a thread state set to XEP_MANAGER_THREAD_STATE_NONE
 *  - for REAL and MAILBOX types: an emtpy list of rule_refs
 *  - for VIRTUAL type: emtpy lists in folder_refs and confition_refs
 *  The rest is O on NULL
 */
XePFolder *
a_Folder_new(XePFolderType type);

/*! Create a copy of a folder
 *  The copy has:
 *  - the same UID, 
 *  - an identical ID
 *  - an identical PATH
 *  - a thread state set to XEP_MANAGER_THREAD_STATE_NONE
 *  - for REAL and MAILBOXES types: a copy of rule_refs
 *  - for VIRTUAL type: a copy of folder_refs and condition_refs
 */ 
XePFolder *
a_Folder_new_copy(XePFolder *folder);

/*! Unregister and free a folder. Must not be used for folders
 *  created with a_Folder_new_copy()
 */
void
a_Folder_free(XePFolder *folder);

/*! Free a folder (without unregistering)
 *  To be used for folders created with a_Folder_new_copy
 */
void
a_Folder_free_copy(XePFolder *folder);

/*! Copy the settings of one source folder to the target folder
 *  The target has:
 *  - an identical ID
 *  - an identitcal path
 *  - for REAL and MAILBOX types: a copy of source's rule_refs (target's
 *    original rule_refs are freeed)
 *  - for VIRTUAL type: a copy of source's folder_refs and condition_refs 
 *    (target's original ones are freeed)
 */
void
a_Folder_copy(XePFolder *source, XePFolder *target);

/*! Find the folder associated to the given Unique Identifier (uid)
 *  \param uid unique identifier
 *  \return the XePFolder associated to it, or NULL if there
 *          is no folder with this uid
 */
XePFolder *
a_Folder_get_from_uid(long uid);

XePFolderType
a_Folder_get_type(XePFolder *folder);

/*! Update folder's ID.
 *  ID is composed from its Mailbox ID and its path.
 *  As setting the mailbox or the path of the folder invalidates its
 *  id string, this function must be called explicitly to recreate it.
 *  \param folder the folder which ID string has to be updated. The
 *         ID string is the concatenation of mailbox's ID and
 *         folder's path with a colon between them.
 */
void
a_Folder_update_id(XePFolder *folder);

/*! Return the folder's ID 
 *  \param folder the folder which ID is sought
 */
const char *
a_Folder_get_id(XePFolder *folder);

/*! Set folder's path, or in case of a VFolder, its name 
 * \param folder the folder which path/name is to be set
 * \param path the path/name of the folder/VFolder
 */
void
a_Folder_set_path(XePFolder *folder, const char *path);

/*! Set folder's path, or in case of a VFolder, its name 
 *  This invalidates folder ID, thus a_Folder_update_id()
 *  should be called afterwards.
 * \param folder the folder which path/name is to be set
 * \param path the path/name of the folder/VFolder, folder
 *        will get ownership on path.
 */
void
a_Folder_set_path_p(XePFolder *folder, char *path);

const char *
a_Folder_get_path(XePFolder *folder);

/*! Set the mailbox to which the folder pertains
 *  This invalidates folder's ID and thus a_Folder_update_id()
 *  should be called afterwards
 */
void
a_Folder_set_mailbox(XePFolder *folder, XePMailbox *mailbox);

XePMailbox *
a_Folder_get_mailbox(XePFolder *folder);

/*! Get a short name for the folder
 *  For VIRTUAL folders, this is its path (i.e. name)
 *  For MAILBOX folders, this is Mailbox ID
 *  For REAL folders:
 *  - IMAP: last part of the path
 *  - MH: idem
 *  - everything else, the complete path
 */
const char *
a_Folder_get_short_name(XePFolder *folder);

void
a_Folder_set_folder_refs(XePFolder *folder, clist *folder_refs);

const clist *
a_Folder_get_folder_refs(XePFolder *folder);

void
a_Folder_set_condition_refs(XePFolder *folder, clist *condition_refs);

const clist *
a_Folder_get_condition_refs(XePFolder *folder);


/*! Create and register a new XePFolderRef
 *  This is a triplet (Ref, XePFolder, enabled) where Ref is
 *  the reference (ID) of a folder, XePFolder contains the pointer
 *  to this referred to folder and enabled is a flag 
 *  XePFolderRefs are registered
 */
XePFolderRef *
a_Folder_ref_new(void);

/*! Create and register a new XePFolderRef with the same reference
 *  as the given one
 */
XePFolderRef *
a_Folder_ref_new_copy(XePFolderRef *folder_ref);

/*! Unregister and release resources used by XePFolderId 
 */
void
a_Folder_ref_free(XePFolderRef *folder_ref);

/*! Release resources used by XePFolderRef content without unregistering it */
void
a_Folder_ref_free_content(XePFolderRef *folder_ref);

/*! Create an empty list that will contain XePFolderRefs 
 */
clist *
a_Folder_ref_list_new(void);

/*! Unregister and free all XePFolderRefs in the list and free the list itself
 */
void
a_Folder_ref_list_free(clist *folder_refs);

/*! Create the global hash table that will store registered folder IDs */
chash *
a_Folder_ref_hash_new(void);

/*! Free the global hash table for XePFolder references
 */
void
a_Folder_ref_hash_free(void);

/*! Return the pointer to the global hash table for XePFolderRefs
 */
chash *
a_Folder_ref_hash_get(void);

/*! Register a folder in the global hash table for folder IDs 
 *  \return 0 on success. Fails if folder's id is already registered
 */
int
a_Folder_register(XePFolder *folder);

/*! Unregister a folder in the global hash table for folder IDs 
 *  \return 0 on success. Fails if folder ID was not found to be registered
 */
int
a_Folder_unregister(XePFolder *folder);

/*! Register a list of folders in the global hash table for folder IDs 
 *  \return 0 on success. Fails as soon as a_Folder_register() fails
 *  for one of the items in the list
 */
int
a_Folder_register_list(clist *folders);

/*! Unregister a list of folders in the global hash table for folder IDs 
 *  \return 0 on success. Fails as soon as a_Folder_unregister() fails
 *  for one of the items in the list
 */
int
a_Folder_unregister_list(clist *folders);

/*! Return the XePFolder registered with the given ID (ref)
 *  \return pointer to the XePFolder if found, NULL otherwise.
 */
XePFolder *
a_Folder_get_from_ref(const char *ref);

/*! Create the rules links used by a folder.
 *  A folder first holds the identifier (XePProcessingRuleId) 
 *  for the processing rules it uses. This function use these
 *  identifiers to search the rules, get the pointer to the
 *  XePProcessingRule and store it. The identifiers are the
 *  freeed and NULLified.
 */
void
a_Folder_update_rules_links(XePFoldersList *folders);

/*! Set the XePFolder member of XePFolderRef to the folder
 *  registered under the reference stored in the ref member
 *  of the XePFolderRef. After successful operation,
 *  the XePFolderRef has a consitent state where its ref
 *  corresponds to its folder
 *  \return 0 on success
 */
int
a_Folder_ref_link_to_folder(XePFolderRef *folder_ref);

/*! Relink all folder references to the correspondig folder
 *  This functions updates the XePFolderRef::folder member
 *  of all known XePFolderRefs.
 */
void
a_Folder_ref_relink_all(void);

/*! Get the state of the thread associated with the folder
 *  \param folder an existing folder 
 *  \return the thread state
 */
XePManagerThreadState
a_Folder_get_thread_state(XePFolder *folder);

int
a_Folder_is_locked(XePFolder *folder);

int
a_Folder_connect(XePFolder *folder);

int
a_Folder_disconnect(XePFolder *folder);

int
a_Folder_noop(XePFolder *folder);

int
a_Folder_expunge(XePFolder *folder);

/*! Sync folder content with mailbox 
 *  Eventually connect to the mailbox and retrieve content.
 *  At the end of the operation, folder should have the same information
 *  as the mailbox.
 */
int
a_Folder_sync(XePFolder *folder);

/*! Checkpoint folder 
 */
int
a_Folder_check(XePFolder *folder);

/*! Append a message to a folder
 *  \param folder
 *  \param data pointer to the rfc2822/822 message
 *  \param size the size of the message
 *  \return MAIL_NO_ERROR on success and MAIL_ERROR_XXX on error
 */
int
a_Folder_append(XePFolder *folder, const char *data, size_t size);

int
a_Folder_mailbox_connect(XePFolder *folder);

int
a_Folder_mailbox_disconnect(XePFolder *folder);

int
a_Folder_mailbox_list(XePFolder *folder, struct mail_list **mb_list);

int
a_Folder_mailbox_lsub(XePFolder *folder, struct mail_list **mb_list);


/*! Append a file to a folder in an asynchronous way then call
 *  the user provided function when finished.
 *  \param folder the target XePFolder
 *  \param filename the file containing the RFC2822/822 message to append
 *  \param user_cb the user callback function called when the operation
 *         finishes
 *  \param user_data the user data to pass to the user callback function
 */
void
a_Folder_append_file_async(XePFolder *folder, const char *filename,
			   XePManagerMsgCb user_cb, void *user_data);

void
a_Folder_disconnect_async(XePFolder *folder,
			  XePManagerMsgCb user_cb, void *user_data);

/*! Get total count of message tree nodes, including dummy messages
 *  this is used for the tree view because it needs to know the 
 *  total number of rows displayed
 */
uint32_t
a_Folder_tree_get_total_count(XePFolder *folder);

void
a_Folder_messages_change_flags_async(XePFolder *folder, 
				     carray *messages,
				     XePManagerMsgCb user_cb,
				     void *user_data);
int
a_Folder_array_append_messages_async(carray *dest_folders, carray *messages,
				     XePFolderAppendMessagesCb user_cb,
				     void *user_data);

XePFolderProcessing *
a_Folder_processing_new(XePFolder *folder, 
			carray *messages,
			clist *processing_nodes);

void
a_Folder_processing_free(XePFolderProcessing *folder_processing);

void
a_Folder_processing(XePFolderProcessing *folder_processing);

clist *
a_Folder_get_processing_rules(XePFolder *folder);

void
a_Folder_processing_rule_add(XePFolder *folder, const char *ref);

void
a_Folder_processing_rule_remove(XePFolder *folder, const char *ref);

#endif
