/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef MAILBOX_H
#define MAILBOX_H

#include <clist.h>

#include "folder-types.h"
#include "mailbox-types.h"
#include "server.h"
#include "manager.h"

typedef void (*XePMailboxListCloneCb)(clist *mailboxes, void *adata);
typedef void (*XePMailboxDisconnectCb)(XePMailbox *mailbox, void *adata);

/*! Mailbox or mail source.
 *  A mail source (mailbox) is defined by its type (POP3, IMAP, NNTP, MBOX or
 *  MH) and a name.
 */
struct XePMailbox_s
{
  char *id;              /**< Unique identifier. FIXME: an integer?  */
  char *name;            /**< Mailbox name, chosen by the user.      */
  long uid;
  XePMailboxType type;   /**< Type (IMAP, MH, MBOX, NNTP, POP3).     */
  XePServer *server;

  char *inbox_ref;
  char *draft_ref;
  char *queue_ref;
  char *sent_ref;

  char *command;         /**< External command to start connection   */
  char *cache;           /**< Cache directory.                       */
  char *flags;           /**< Flags directory.                       */

  int use_multiple_connections; /**< Wether to use a separate connection 
				 *   for each folder
				 */
  int max_connections;          /**< If mutlitple connections allowed,
				 *   limit their number to max_connections
				 */

  int subscribed_only;
  XePFolder *folder;            /**< The folder representing the mailbox
				 *   it also holds the root path, if needed
				 */
  clist *folders;
  struct mail_list *ep_mail_list;
  struct mail_list *ep_mail_lsub;
};

/*! Create an empty list to hold XePMailboxes 
 */
clist *
a_Mailbox_list_new(void);

/*! Create a list of mailboxes that are copies of the given list of
 *  mailboxes. Copies are done with a_Mailbox_new_copy()
 *  \return the list of copies of success, NULL otherwise
 */
clist *
a_Mailbox_list_new_copy(clist *mailboxes);

/*! Free the mailboxes in the given list and free the list itself afterwards
 *  Mailboxes are freeed with a_Mailbox_free()
 */
void
a_Mailbox_list_free(clist *mailboxes);

/*! Free the mailbox copies in the given list and free the list itself 
 *  aftewards. Mailboxes are freeed with a_Mailbox_free_copy()
 */
void
a_Mailbox_list_free_copy(clist *mailboxes);

/*! Create a usable list of mailboxes from a list of mailbox copies
 *  The given list consist of a list of mailboxes created with 
 *  a_Mailbox_list_new_copy(). These mailbox copies have the 
 *  same uid as the original ones, but they are not registered.
 *  Each mailbox in the list which has a negative uid will be freed
 *  When finished, it calls the user_cb() function with the new
 *  mailbox list as first argument and the user_data as second argument
 *  \param mailboxes_copy list of mailboxes copies
 *  \param user_cb a user provided function called when operation finishes
 *  \para user_data the user data to pass as the second argument to user_cb()
 *  \return 0 on success and -1 if malloc() fails
 *  This is an asynchronous operation. The result is given to user_cb()
 *  in its first argument
 */
int
a_Mailbox_list_clone(clist *mailboxes_copy,
		     XePMailboxListCloneCb user_cb,
		     void *user_data);

/*! Create and register a new mailbox
 *  The newly created and registered mailbox has:
 *  - a strictly positive unique identifier
 *  - a MAILBOX folder which is NOT registered (MAILBOX folder are
 *    registered only for mbox and POP3 mailboxes)
 *  - an UNKNOWN type
 *  - USE_MULTIPLE_CONNECTIONS set
 *  - max_connections set to XEP_MAILBOX_MAX_CONNECTIONS
 *  - an empty folders list
 *  the rest of members are 0 or NULL
 */
XePMailbox *
a_Mailbox_new(const char *id);

/*! Create but do not register a copy of a mailbox
 *  The newly created mailbox has:
 *  - an identical name
 *  - the same UID
 *  - the same type
 *  - an identical server
 *  - identicals defaults folders references
 *  - identicals cache and flags settings
 *  - identical command
 *  - identical subscribed_only
 *  - a copy of MAILBOX folder created with a_Folder_new_copy() that 
 *    has its mailbox member pointing to this mailbox copy
 *  - a copy of the list of folders created with a_Folder_list_new_copy()
 */
XePMailbox *
a_Mailbox_new_copy(XePMailbox *mailbox);

/*! Unregister and free a mailbox
 *  MAILBOX folder is freeed using a_Folder_free()
 *  folders list is freeed using a_Folder_list_free()
 */
void
a_Mailbox_free(XePMailbox *mailbox);

/*! Free a mailbox copy
 *  MAILBOX folder is freeed using a_Folder_free_copy()
 *  folders list is freeed using a_Folder_list_free_copy()
 */
void
a_Mailbox_free_copy(XePMailbox *mailbox);

/*! Set the type of the mailbox
 *  For MBOX and POP3 mailboxes, the MAILBOX folder
 *  is promoted to a REAL type. For POP3 mailboxes,
 *  the previous MAILBOX folder is unregistered if previous type
 *  was also POP3 or if previous MAILBOX folder had a path, then
 *  it is registered. 
 */
void
a_Mailbox_set_type(XePMailbox *mailbox, XePMailboxType type);

/*! Set the toplevel directory of a mailbox
 *  If the MAILBOX folder had already a path or if mailbox type is POP3,
 *  unregister of MAILBOX folder, update the path and register again
 *  with new toplevel (path)
 *  This means, when toplevel is set or mailbox is of type POP3,
 *  then MAILBOX folder is registered  
 */
void
a_Mailbox_set_toplevel(XePMailbox *mailbox, const char *toplevel);

const char *
a_Mailbox_get_id(XePMailbox *mailbox);

XePMailbox *
a_Mailbox_get_from_uid(long uid);

const char *
a_Mailbox_type_as_string(XePMailboxType type);

XePMailboxType
a_Mailbox_type_from_string(const char *type);

/*! Check whether mailbox folders are running 
 *  They are considered running if their thread state is
 *  XEP_MANAGER_THREAD_STATE_RUNNING
 *  \return -1 if there is at least one folder running, 0 other wise
 */
int
a_Mailbox_is_busy(XePMailbox *mailbox);

/*! Disconnect all connections pertaining to the mailbox then
 *  call the user provided function.
 *  This function calls the user provided function immediatly
 *  if no connections exist. Otherwise, the user callback
 *  will be called by the main loop when all connections have been
 *  shut down.
 */
void
a_Mailbox_disconnect_all(XePMailbox *mailbox,
			 XePMailboxDisconnectCb user_cb,
			 void *user_data);
/*! Disconnect a mailbox then call the user provided function.
 *  If the mailbox is already disconnected, this function
 *  calls the user callback immediatly before returning. Otherwise
 *  it returns and the user provided callback is called by the main
 *  loop when the mailbox has been disconnected.
 */
void
a_Mailbox_disconnect_async(XePMailbox *mailbox,
 			   XePManagerMsgCb user_cb, 
			   void *user_data);
#endif
