/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <errno.h>
#include <fcntl.h> /* open */
#include <string.h>
#include <sys/types.h> /* open */
#include <sys/stat.h> /* open */
#include <time.h>
#include <unistd.h> /* close */

#include <clist.h>
#include <charconv.h>
#include <libetpan.h>

#include "intl.h"

#include "logging.h"
#include "x_alloc.h"

#include "utils.h"

#include "message.h"
#include "folder.h"

#define DATE_BUF_MAX 64


XePMessageHeader message_headers[] =
  {
    {XEP_MAIL_HEADER_ORIG_DATE,   N_("Date: "),        MAILIMF_FIELD_ORIG_DATE},
    {XEP_MAIL_HEADER_FROM,        N_("From: "),        MAILIMF_FIELD_FROM},
    {XEP_MAIL_HEADER_SENDER,      N_("Sender: "),      MAILIMF_FIELD_SENDER},
    {XEP_MAIL_HEADER_REPLY_TO,    N_("Reply-To: "),    MAILIMF_FIELD_REPLY_TO},
    {XEP_MAIL_HEADER_TO,          N_("To: "),          MAILIMF_FIELD_TO},
    {XEP_MAIL_HEADER_CC,          N_("CC: "),          MAILIMF_FIELD_CC},
    {XEP_MAIL_HEADER_BCC,         N_("BCC: "),         MAILIMF_FIELD_BCC},
    {XEP_MAIL_HEADER_MESSAGE_ID,  N_("Message-ID: "),  MAILIMF_FIELD_MESSAGE_ID},
    {XEP_MAIL_HEADER_IN_REPLY_TO, N_("In-Reply-To: "), MAILIMF_FIELD_IN_REPLY_TO},
    {XEP_MAIL_HEADER_REFERENCES,  N_("References: "),  MAILIMF_FIELD_REFERENCES},
    {XEP_MAIL_HEADER_SUBJECT,     N_("Subject: "),     MAILIMF_FIELD_SUBJECT},
    {XEP_MAIL_HEADER_COMMENTS,    N_("Comments: "),    MAILIMF_FIELD_COMMENTS},
    {XEP_MAIL_HEADER_KEYWORDS,    N_("Keywords: "),    MAILIMF_FIELD_KEYWORDS},
    {XEP_MAIL_HEADER_OTHER,       NULL,                MAILIMF_FIELD_OPTIONAL_FIELD} /* Sentinel: this must be the last */
  };

static void
Message_reset(XePMessage *message);

static char *
Message_message_formatted_addr_list(clist *list, char *template,
				    XePUtilsFormat *format,
				    char *dest_charset);
static int
Message_fetch_part_decoded(mailmessage *msg, 
			   struct mailmime *mime,
			   struct mailmime_single_fields *fields,
			   char **data,
			   size_t *len);

static struct mailmime *
Message_mime_find_default(mailmessage *msg,
			  struct mailmime *mime);

static struct mailimf_mailbox *
Message_mailbox_clone(struct mailimf_mailbox *mb);

static struct mailimf_address_list *
Message_address_list_new_from_mailbox_list(struct mailimf_mailbox_list *mb_list);

static int
Message_create_in_reply_to(struct mailimf_single_fields *single_fields,
			   clist **result);

static int
Message_create_references(struct mailimf_single_fields *single_fields,
			  clist **references);

static void
Message_set_user_agent(struct mailimf_fields *fields);

static int
Message_add_optional_header(struct mailimf_fields *fields,
			    const char *header_name,
			    const char *header_value);

static int
Message_get_subject(struct mailimf_single_fields *single_fields,
		    char **subject);

/* Function definitions */

/* Message creation and deletion functions */

/*! Must be called from the main thread */
XePMessage *
a_Message_new(void)
{
  XePMessage *message;
  /* Start muid from 1, 0 meaning no message being displayed. 
   * No guarantees are given on what happens for the 2^32 
   * displayed message (in the same session).
   */
  static uint32_t muid = 1; 

  message = x_alloc(1, XePMessage);
  if (message == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      message->data = NULL;
      message->len = 0;
      message->muid = muid++;
      message->async = 0;
      message->folder = NULL;
      message->msg = NULL;
      message->mime = NULL;
      message->fields = NULL;
      message->single_fields = NULL;
      message->selected_mime = NULL;
      message->default_mime = NULL;
      message->source_charset = NULL;
      message->dest_charset = NULL;
      message->flags_set = 0;
      message->flags_unset = 0;
      message->owner = 0;
      message->filename = NULL;
      message->fd = -1;
      message->msg_data = NULL;
      message->msg_size = 0;
    }

  return message;
}

XePMessage *
a_Message_new_from_data_noparse(char *data, size_t size)
{
  XePMessage *message;

  message = a_Message_new();
  if (message)
    {
      message->msg_data = data;
      message->msg_size = size;
      message->owner = 1;
    }

  return message;
}

XePMessage *
a_Message_new_from_file_noparse(const char *filename)
{
  XePMessage *message;

  message = a_Message_new();
  if (message)
    {
      message->filename = strdup(filename);
      message->owner = 1;
      if (message->filename == NULL)
	{
	  A_LOGGING_ERROR("strdup(): %s", strerror(errno));
	  a_Message_free(message);
	  message = NULL;
	}
    }

  return message;
}

XePMessage *
a_Message_new_from_data(char *data, size_t len)
{
  XePMessage *message = NULL;

  struct mailimf_fields *imf_fields = NULL;
  struct mailmessage *msg;
  struct mailmime *mime;
  size_t index = 0;
  int r;

  r = mailimf_fields_parse(data, len, &index, &imf_fields);
  if (r != MAILIMF_NO_ERROR)
    {
      A_LOGGING_ERROR("Error while parsing message");
    }
  else
    {
      int cur_token = 0;
      msg = mailmessage_new();
      msg->fields = imf_fields;
      mailmime_parse(data, len, &cur_token, &mime);

      mailmessage_resolve_single_fields(msg);
      
      message = a_Message_new();
      message->msg = msg;
      message->mime = mime;
      message->msg_data = data;
      message->owner = 1;
    }

  return message;
}

XePMessage *
a_Message_new_for_fetching(XePFolder *folder, mailmessage *msg)
{
  XePMessage *message;

  message = a_Message_new();
  if (message)
    {
      message->async = 1;
      message->folder = folder;
      message->msg = msg;
    }

  return message;
}

XePMessage *
a_Message_new_to_compose(void)
{
  XePMessage *message;
  mailmessage *msg;
  struct mailmime *mime;
  struct mailimf_fields *fields;
  int is_nntp = 0;

  message = a_Message_new();
 
  message->owner = 1;
  
  msg = mailmessage_new();

  message->msg = msg;

  fields = mailimf_fields_new_with_data(NULL,
					NULL, /* Sender:   */
					NULL, /* Reply-To: */
					NULL, /* To        */
					NULL, /* Cc        */
					NULL, /* Bcc       */
					NULL,
					NULL,
					NULL);

  Message_set_user_agent(fields);

  mime = a_Message_mime_new_message("rfc822", MAILMIME_MECHANISM_8BIT);
  mailmime_set_imf_fields(mime, fields);
  mailmime_set_body_text(mime, NULL, 0);
  mailimf_single_fields_init(&msg->single_fields, fields);

  message->mime = mime;
  message->selected_mime = mime;

  mime = a_Message_mime_add_part_to_selected(message, NULL, 0, 
					     "text/plain",
					     MAILMIME_MECHANISM_8BIT);
  message->default_mime = mime;
  message->selected_mime = mime;

  
  

  message->source_charset = NULL;

  return message;
}

XePMessage *
a_Message_new_to_reply(XePMessage *message)
{
  XePMessage *reply_message;
  mailmessage *reply_msg;
  struct mailmime *mime;
  struct mailimf_fields *reply_fields;
  struct mailimf_mailbox_list * from = NULL;
  struct mailimf_address_list *to = NULL;
  struct mailimf_address_list *cc;
  clist *in_reply_to = NULL;
  clist *references = NULL;
  char *subject = NULL;
  char *reply_subject = NULL;
  struct mailimf_single_fields *single_fields;
  const char *reply_prefix = "Re: ";
  int is_nntp = 0;

  a_Message_get_bodystructure(message);

  if ((message->msg == NULL) || (message->msg->resolved == 0))
    return NULL;

  single_fields = &(message->msg->single_fields);

  reply_message = a_Message_new();
  reply_message->owner = 1;
  
  reply_msg = mailmessage_new();
  reply_message->msg = reply_msg;

  /* To: */

  if (single_fields->from)
    {
      to = Message_address_list_new_from_mailbox_list(single_fields->from->mb_list);
    }

  Message_create_in_reply_to(single_fields, &in_reply_to);
  Message_create_references(single_fields, &references);
  Message_get_subject(single_fields, &subject);

#if 0
  /* TODO: reply subject created from a s/// at user preference level */
  reply_subject = a_Prefs_reply_subject_new(subject);
#else
  if (subject)
    {
      reply_subject = x_alloc(strlen(subject) + strlen(reply_prefix) + 1, char);
      strcpy(reply_subject, reply_prefix);
      strcat(reply_subject, subject);
    }
#endif

  reply_fields = mailimf_fields_new_with_data(from,
					      NULL, /* Sender:   */
					      NULL, /* Reply-To: */
					      to,   /* To        */
					      NULL, /* Cc        */
					      NULL, /* Bcc       */
					      in_reply_to,
					      references,
					      reply_subject);
#if 0
  /* TODO: set newsgroups header */
  if (is_nntp)
    {
      char *newsgroups;
      /* msg->fields has to be fetched with mailmessage_fetch_header()? */
      Message_get_followup_to(msg->fields, &newsgroups);
      Message_set_newsgroup(newsgroups);
    }
#endif
  
  Message_set_user_agent(reply_fields);

  mime = a_Message_mime_new_message("rfc822", MAILMIME_MECHANISM_8BIT);
  mailmime_set_imf_fields(mime, reply_fields);
  mailmime_set_body_text(mime, NULL, 0);
  mailimf_single_fields_init(&reply_msg->single_fields, reply_fields);

  reply_message->mime = mime;
  reply_message->selected_mime = mime;

  mime = a_Message_mime_add_part_to_selected(reply_message, NULL, 0, 
					     "text/plain",
					     MAILMIME_MECHANISM_8BIT);

  reply_message->default_mime = mime;
  reply_message->selected_mime = mime;

  /* TODO: quoting depend on text analysis and hence on 
   * character set
   */
  reply_message->source_charset = (message->source_charset ? 
				   strdup(message->source_charset) :
				   NULL);

  return reply_message;
}

XePMessage *
a_Message_new_to_forward(XePMessage *message)
{
  XePMessage *forward_message;
  mailmessage *forward_msg;
  struct mailimf_fields *forward_fields;
  struct mailimf_address_list *to;
  clist *in_reply_to;
  clist *references;

  a_Message_get_bodystructure(message);
#if 0
  if ((message->msg == NULL) || (message->msg->single_fields) == NULL)
    return NULL;
#endif
  forward_message = a_Message_new();
 
  forward_message->owner = 1;
  
  forward_msg = mailmessage_new();

  forward_message->msg = forward_msg;

  A_LOGGING_SYSTEM_ERROR("Incomplete");
  return forward_message;
}

XePMessage *
a_Message_new_to_redirect(XePMessage *message)
{
  XePMessage *redirect_message = NULL;
  mailmessage *redirect_msg;
  struct mailimf_fields *redirect_fields;
  struct mailimf_address_list *to;
  clist *in_reply_to;
  clist *references;

  if (a_Message_get_bodystructure(message))
    return NULL;

  A_LOGGING_SYSTEM_ERROR("Incomplete");
  return redirect_message;
}

static void
Message_reset(XePMessage *message)
{

  if (message->owner)
    {
      if (message->msg)
	{
	  mailmessage_free(message->msg);
	}
      if (message->mime)
	{
	  mailmime_free(message->mime);
	}
      x_free(message->msg_data);
      message->msg_size = 0;
      x_free(message->filename);
      message->msg_data = NULL;
      message->filename = NULL;
      a_Message_file_close(message);
    }

  x_free(message->data);
  x_free(message->source_charset);
  message->data = NULL;
  message->len = 0;
  message->msg = NULL;
  message->mime = NULL;
  message->fields = NULL;
  message->single_fields = NULL;
  message->default_mime = NULL;
  message->selected_mime = NULL;
  
}

void
a_Message_free(XePMessage *message)
{
  if (message)
    {
      Message_reset(message);
      x_free(message);
    }
}

void
a_Message_free_cb(void *amessage, void *anargout)
{
  a_Message_free((XePMessage *) amessage);
}

XePMessage*
a_Message_clone(XePMessage *message)
{
  XePMessage *clone = NULL;
  if (message)
    {
      clone = a_Message_new();
      if (clone)
	{
#if 0
	  memcpy(clone, message, sizeof(message));
#else
	  *clone = *message;
#endif

	}
    }
  return clone;
}

int
a_Message_file_open(XePMessage *message)
{
  int r = 0;
  if (message->filename)
    {
      message->fd = open(message->filename, O_RDONLY);
      if (message->fd == -1)
	{
	  A_LOGGING_ERROR("open(): %s", strerror(errno));
	  r = -1;
	}
    }
  else
    {
      A_LOGGING_ERROR("Cannot open message file: message not linked to a file");
      r = -1;
    }
  return r;
}

void
a_Message_file_close(XePMessage *message)
{
  if (message->fd >= 0)
    {
      close(message->fd);
      message->fd = -1;
    }
}

void
a_Message_set_folder(XePMessage *message, XePFolder *folder)
{
  message->folder = folder;
}

XePFolder *
a_Message_get_folder(XePMessage *message)
{
  return message->folder;
}

/* Functions to create message headers */

static int
Message_create_references(struct mailimf_single_fields *single_fields,
			  clist **references)
{
  clist *result = NULL;
  char *refid;
  char *msgid;
  clistiter *iter;
  int r = 0;

  result = clist_new();
  if (result == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new() failed");
      return -1;
    }
  if (single_fields->references)
    {
      for (iter = clist_begin(single_fields->references->msg_id_list);
	   iter && (r == 0);
	   iter = iter->next)
	{
	  if (iter->data)
	    {
	      refid = strdup(((char *) iter->data));
	      if (refid)
		{
		  r = clist_append(result, refid);
		  if (r)
		    {
		      free(refid);
		    }
		}
	      else
		{
		  A_LOGGING_SYSTEM_ERROR("strdup() failed");
		  r = -1;
		}
	    }
	}
    }
  if (r == 0 && single_fields->message_id)
    {
      msgid = strdup((single_fields->message_id->value));
      if (msgid)
	{
	  r = clist_append(result, msgid);
	  if (r)
	    {
	      free(msgid);
	    }
	}
    }

  if (r)
    {
      clist_foreach(result, (clist_func) free, NULL);
      clist_free(result);
      result = NULL;
    }

  *references = result;
  return r;
}

static int
Message_create_in_reply_to(struct mailimf_single_fields *single_fields,
			   clist **result)
{
  clist *in_reply_to = NULL;
  char *msgid;
  char *msgid_dup;
  int r = 0;

  if (single_fields->message_id)
    {
      msgid = single_fields->message_id->value;
      if (msgid)
	{
	  msgid_dup = strdup(msgid);
	  if (msgid_dup)
	    {
	      in_reply_to = clist_new();
	      if (in_reply_to)
		{
		  if (clist_append(in_reply_to, msgid_dup))
		    {
		      r = -1;
		      clist_foreach(in_reply_to, (clist_func) free, NULL);
		      clist_free(in_reply_to);
		      in_reply_to = NULL;
		      x_free(msgid_dup);
		      msgid_dup = NULL;
		    }
		}
	      else
		{
		  r = -1;
		  x_free(msgid_dup);
		  msgid_dup = NULL;
		}		  
	    }
	}
    }

  *result = in_reply_to;

  return r; 
}

/* Email addresses functions */

static struct mailimf_address_list *
Message_header_get_addr_list(struct mailimf_fields *fields, int field_type)
{
  clistiter *iter;
  struct mailimf_field *field;
  struct mailimf_bcc      *bcc        = NULL;
  struct mailimf_cc       *cc         = NULL;
  struct mailimf_reply_to *reply_to   = NULL;
  struct mailimf_bcc      *resent_bcc = NULL;
  struct mailimf_cc       *resent_cc  = NULL;
  struct mailimf_to       *resent_to  = NULL;
  struct mailimf_to       *to         = NULL;
  struct mailimf_address_list *addr_list;
  clist *list;

  for (iter = clist_begin(fields->list); iter; iter = iter->next)
    {
      field = (struct mailimf_field *) iter->data;
      if (field->type == field_type)
	{
	  switch (field_type)
	    {
	    case MAILIMF_FIELD_BCC:
	      return field->field.bcc->addr_list;
	      break;
	    case MAILIMF_FIELD_CC:
	      return field->field.cc->addr_list;
	      break;
	    case MAILIMF_FIELD_REPLY_TO:
	      return field->field.reply_to->addr_list;
	      break;
	    case MAILIMF_FIELD_RESENT_BCC:
	      return field->field.resent_bcc->addr_list;
	      break;
	    case MAILIMF_FIELD_RESENT_CC:
	      return field->field.resent_cc->addr_list;
	      break;
	    case MAILIMF_FIELD_RESENT_TO:
	      return field->field.resent_to->addr_list;
	      break;
	    case MAILIMF_FIELD_TO:
	      return field->field.to->addr_list;
	      break;
	    default:
	      A_LOGGING_SYSTEM_ERROR("Illegal field type in this context");
	      return NULL;
	    }
	}
    }
  
  list = clist_new();

  addr_list = mailimf_address_list_new(list);

  switch (field_type)
    {
    case MAILIMF_FIELD_BCC:
      bcc = mailimf_bcc_new(addr_list);
      break;
    case MAILIMF_FIELD_CC:
      cc = mailimf_cc_new(addr_list);
      break;
    case MAILIMF_FIELD_REPLY_TO:
      reply_to = mailimf_reply_to_new(addr_list);
      break;
    case MAILIMF_FIELD_RESENT_BCC:
      resent_bcc = mailimf_bcc_new(addr_list);
      break;
    case MAILIMF_FIELD_RESENT_CC:
      resent_cc = mailimf_cc_new(addr_list);
      break;
    case MAILIMF_FIELD_RESENT_TO:
      resent_to = mailimf_to_new(addr_list);
      break;
    case MAILIMF_FIELD_TO:
      to = mailimf_to_new(addr_list);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Illegal field type in this context");
      mailimf_address_list_free(addr_list);
      return NULL;
    }

  field = mailimf_field_new(field_type,
			    NULL,      NULL,       NULL, NULL, resent_to,
			    resent_cc, resent_bcc, NULL, NULL, NULL,
			    NULL,      reply_to,   to,   cc,   bcc,
			    NULL,      NULL,       NULL, NULL, NULL,
			    NULL,      NULL);
  clist_append(fields->list, field);
  
  return addr_list;
}

static struct mailimf_address_list *
Message_address_list_new_from_mailbox_list(struct mailimf_mailbox_list *mb_list)
{
  struct mailimf_address_list *addr_list;
  struct mailimf_address *addr;
  struct mailimf_mailbox *mb;
  struct mailimf_mailbox *clone_mb;
  
  clistiter *iter;
  clist *list;
  int ok = 1;

  list = clist_new();
  if (list == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new() failed");
      return NULL;
    }

  for (iter = clist_begin(mb_list->list); iter && ok; iter = iter->next)
    {
      mb = (struct mailimf_mailbox *) iter->data;
      clone_mb = Message_mailbox_clone(mb);
      ok = 0;
      if (clone_mb)
	{
	  addr = mailimf_address_new(MAILIMF_ADDRESS_MAILBOX, clone_mb, NULL);
	  if (addr)
	    {
	      if (clist_append(list, addr) == 0)
		{
		  ok = 1;
		}
	      else
		{
		  A_LOGGING_SYSTEM_ERROR("clist_append(): failed");
		  mailimf_address_free(addr);
		}
	    }
	}
    }
  
  if (ok)
    {
      addr_list = mailimf_address_list_new(list);
    }
  else
    {
      clist_foreach(list, (clist_func) mailimf_address_free, NULL);
      clist_free(list);
      addr_list = NULL;
    }

  return addr_list;
}

static int
Message_address_list_add(struct mailimf_address_list *addr_list,
			 const char *address,
			 const char *charset)
{
  struct mailimf_address *addr;
  struct mailimf_mailbox *mb = NULL;
  int cur_token = 0;
  int r = 0;

  A_LOGGING_TRACE;

  while (cur_token < strlen(address) && r == 0)
  {  
    r = mailimf_mailbox_parse((char *)address, strlen(address),
			      &cur_token, &mb);

    if (mb->display_name)
      {
	char *qp;
	if (a_Utils_has_8bit(mb->display_name, 0))
	  {
	    qp = a_Utils_encode_8bit_header(mb->display_name, 
					    charset, 1, 1, 0);
	    if (qp)
	      {
		free(mb->display_name);
		mb->display_name = qp;
	      }
	  }
      }
    
    if (r == MAILIMF_NO_ERROR)
      {
	addr = mailimf_address_new(MAILIMF_ADDRESS_MAILBOX, mb, NULL);
    
	r = clist_append(addr_list->list, addr);
	if (r)
	  {
	    A_LOGGING_SYSTEM_ERROR("clist_append() failed");
	    mailimf_address_free(addr);
	  }
      }
    cur_token++;
  }

  return r;
}

static char *
Message_message_formatted_addr_list(clist *list, char *template,
				    XePUtilsFormat *format,
				    char *dest_charset)
{
  clistiter *address_iter;
  struct mailimf_address *address;
  char *addr = NULL;
  char *result = NULL;
  char *decoded = NULL;
  const char *sep = ", ";
  int first = 1;
  int r;
  int cur_token;
  
  for (address_iter = clist_begin(list);
       address_iter; 
       address_iter = address_iter->next)
    {
      address = (struct mailimf_address *) address_iter->data;
      if (address->type == MAILIMF_ADDRESS_MAILBOX)
	{
	  format[0].string = (address->mailbox->display_name ? 
			      address->mailbox->display_name :
			      address->mailbox->addr_spec);
	  format[1].string = address->mailbox->addr_spec;
	  addr = a_Utils_printf(template, format);

	  result = a_Utils_append(result, (first ? "" : sep), addr, NULL);
	  first = 0;      
	  x_free(addr);
	}
      else if (address->type == MAILIMF_ADDRESS_GROUP)
	{
	  clistiter *group_iter;
	  struct mailimf_mailbox *group_mailbox;
	  
	  result = a_Utils_append(result, (first ? "" : sep),
				  address->group->display_name,
				  "(", NULL);
	  first = 1;
	  
	  for (group_iter = clist_begin(address->group->mb_list->list);
	       group_iter; group_iter = group_iter->next)
	    {
	      group_mailbox = (struct mailimf_mailbox *) group_iter->data;
	      format[0].string = (group_mailbox->display_name ?
				  group_mailbox->display_name :
				  group_mailbox->addr_spec);
	      format[1].string = group_mailbox->addr_spec;
	      addr = a_Utils_printf(template, format);
	      
	      result = a_Utils_append(result, (first ? "" : sep), addr, NULL);
	      first = 0;	      
	      x_free(addr);
	    }
	  result = a_Utils_append(result, ")", NULL);
	  first = 0;
	}
    }

  /* TODO: convert! */
  cur_token = 0;
  r = mailmime_encoded_phrase_parse(DEFAULT_SOURCE_CHARSET,
				    result,
				    strlen(result),
				    &cur_token,
				    dest_charset,
				    &decoded);
  if (r == MAILIMF_NO_ERROR)
    {
      x_free(result);
    }
  else
    {
      decoded = result;
    }

  return decoded;
}

static struct mailimf_mailbox *
Message_mailbox_clone(struct mailimf_mailbox *mb)
{
  struct mailimf_mailbox *clone_mb;
  char *display_name = NULL;
  char *addr_spec = NULL;

  if (mb->display_name)
    {
      display_name = strdup(mb->display_name);
      if (display_name == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("strdup() failed");
	  return NULL;
	}
    }

  if (mb->addr_spec)
    {
      addr_spec = strdup(mb->addr_spec);
      if (addr_spec == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("strdup() failed");
	  x_free(addr_spec);
	  return NULL;
	}
    }

  clone_mb = mailimf_mailbox_new(display_name, addr_spec);
  if (clone_mb == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("mailimf_mailbox_new() failed");
      x_free(display_name);
      x_free(addr_spec);
      return NULL;
    }

  return clone_mb;  
}

/* Functions to convert between the string and the internal representation 
 * of headers
 */

const char *
a_Message_header_get_name_from_type(XePMessageHeaderType type)
{
  return message_headers[type].name;
}

XePMessageHeaderType 
a_Message_header_get_type_from_name(const char *header_name)
{
  XePMessageHeader *header;

  for (header = message_headers; 
       header->name && strcasecmp(header->name, header_name); header++)
    {
      /* Empty on purpose: the search is done in the for conditional */
    }
  return header->type;
}

/* Generic functions to set and get header values */

struct mailimf_field *
a_Message_get_field(XePMessage *message,
		    XePMessageHeaderType type,
		    const char *other_header)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_field *field;
  int ep_type = message_headers[type].ep_type;
  clistiter *iter;

  iter = clist_begin(fields->list);
  while (iter)
    {
      field = (struct mailimf_field *) iter->data;
      if ((field->type == ep_type) &&
	  ((field->type != MAILIMF_FIELD_OPTIONAL_FIELD) ||
	   (strcmp(field->field.optional_field->name, other_header) == 0)))
	{
	  return field;
	}
      else	  
	{
	  iter = iter->next;
	}
    }
  return NULL;
}

			     

void
a_Message_header_add_by_name(XePMessage *message,
			     const char *header_name,
			     const char *header_value,
			     const char *header_charset)
{
  XePMessageHeaderType type;
  
  type = a_Message_header_get_type_from_name(header_name);
  a_Message_header_add(message, type, header_name, header_value,
		       header_charset);
}

void
a_Message_header_add(XePMessage *message,
		     XePMessageHeaderType type,
		     const char *header_name,
		     const char *header_value,
		     const char *header_charset)
{
  switch (type)
    {
    case XEP_MAIL_HEADER_FROM:
      a_Message_header_set_from(message, header_value, header_charset);
      break;
    case XEP_MAIL_HEADER_SUBJECT:
      a_Message_header_set_subject(message, header_value, header_charset);
      break;
    case XEP_MAIL_HEADER_TO:
      a_Message_header_add_to(message, header_value, header_charset);
      break;
    case XEP_MAIL_HEADER_CC:
      a_Message_header_add_cc(message, header_value, header_charset);
      break;
    case XEP_MAIL_HEADER_BCC:
      a_Message_header_add_bcc(message, header_value, header_charset);
      break;
    case XEP_MAIL_HEADER_REPLY_TO:
      a_Message_header_add_reply_to(message, header_value, header_charset);
      break;
    case XEP_MAIL_HEADER_OTHER:
      a_Message_header_add_other(message, header_name, header_value,
				 header_charset);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Unknown header type");
    }
}
		     

void
a_Message_header_remove_by_name(XePMessage *message, const char *header_name)
{
  XePMessageHeaderType type;

  type = a_Message_header_get_type_from_name(header_name);
  a_Message_header_remove(message, type, header_name);
}

void
a_Message_header_remove(XePMessage *message,
			XePMessageHeaderType type,
			const char *other_header)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_field *field;
  int ep_type = message_headers[type].ep_type;
  clistiter *iter;

  iter = clist_begin(fields->list);
  while (iter)
    {
      field = (struct mailimf_field *) iter->data;
      if ((field->type == ep_type) &&
	  ((field->type != MAILIMF_FIELD_OPTIONAL_FIELD) ||
	   (strcmp(field->field.optional_field->name, other_header) == 0)))
	{
	  clist_delete(fields->list, iter);
	  mailimf_field_free(field);
	  iter = clist_begin(fields->list); /* We could optimize this knowing
					     * how clist delete things
					     */
	}
      else	  
	{
	  iter = iter->next;
	}
    }
}

char *
a_Message_message_formatted_get_header(mailmessage *msg, 
				       XePMessageHeaderType type,
				       char *other_header,
				       char *template,
				       char *dest_charset)
{
  char *value = NULL;
  struct mailimf_single_fields *single_fields = &(msg->single_fields);

  switch (type)
    {
    case XEP_MAIL_HEADER_ORIG_DATE:
      value = a_Message_message_formatted_get_local_date(single_fields, 
							 template,
							 dest_charset);
      break;
    case XEP_MAIL_HEADER_FROM:
      value = a_Message_message_formatted_get_from(single_fields,
						   template,
						   dest_charset);
      break;
    case XEP_MAIL_HEADER_SENDER:
      value = a_Message_message_formatted_get_sender(single_fields,
						  template,
						  dest_charset);
      break;
    case XEP_MAIL_HEADER_REPLY_TO:
      value = a_Message_message_formatted_get_reply_to(single_fields,
						    template,
						    dest_charset);
      break;
    case XEP_MAIL_HEADER_TO:
      value = a_Message_message_formatted_get_to(single_fields,
						 template,
						 dest_charset);
      break;
    case XEP_MAIL_HEADER_CC:
      value = a_Message_message_formatted_get_cc(single_fields,
						 template,
						 dest_charset);
      break;
    case XEP_MAIL_HEADER_BCC:
      value = a_Message_message_formatted_get_bcc(single_fields,
						  template,
						  dest_charset);
      break; 
    case XEP_MAIL_HEADER_MESSAGE_ID:
      value = a_Message_message_formatted_get_message_id(single_fields,
							 template,
							 dest_charset);
      break;
    case XEP_MAIL_HEADER_IN_REPLY_TO:
      value = a_Message_message_formatted_get_in_reply_to(single_fields,
							  template,
							  dest_charset);
      break;
    case XEP_MAIL_HEADER_REFERENCES:
      value = a_Message_message_formatted_get_references(single_fields,
							 template,
							 dest_charset);
      break;
    case XEP_MAIL_HEADER_SUBJECT:
      a_Message_message_get_subject(single_fields,
				    dest_charset,
				    &value);
      break;
    case XEP_MAIL_HEADER_COMMENTS:
      value = a_Message_message_formatted_get_comments(single_fields,
						       template,
						       dest_charset);
      break;
    case XEP_MAIL_HEADER_KEYWORDS:
      value = a_Message_message_formatted_get_keywords(single_fields,
						       template,
						       dest_charset);
      break;
    case XEP_MAIL_HEADER_OTHER:
      a_Message_message_get_other(msg->fields,
				  other_header, 
				  dest_charset,
				  &value);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Unknown header type");
    }
  return value;
}

/* Subject: header functions */

static int
Message_get_subject(struct mailimf_single_fields *single_fields,
		    char **subject)
{
  int r = 0;
  if (single_fields->subject == NULL)
    {
      *subject = NULL;
    }
  else
    {
      *subject = strdup((single_fields->subject->value));
      if (*subject == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("strdup() failed");
	  r = -1;
	}
    }
  return r;
}

int
a_Message_message_get_subject(struct mailimf_single_fields *fields,
			      char *dest_charset,
			      char **subject)
{
  int r;
  size_t cur_token;
  char *original_subject;

  A_LOGGING_TRACE;

  if (fields->subject == NULL)
    return -1;

  A_LOGGING_TRACE;

  original_subject = fields->subject->value;

  cur_token = 0;
  r = mailmime_encoded_phrase_parse(DEFAULT_SOURCE_CHARSET,
				    original_subject,
				    strlen(original_subject),
				    &cur_token,
				    dest_charset,
				    subject);
  if (r != MAILIMF_NO_ERROR)
    {
      *subject = strdup(original_subject);
    }
		
  a_Utils_string_unfold(*subject);

  return 0;
}

int
a_Message_header_set_subject(XePMessage *message, 
			     const char *header_value, 
			     const char *header_charset)
{
  char *subject;
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_subject *subject_field;
  struct mailimf_field *field;
  int r = 0;

  a_Message_header_remove(message, XEP_MAIL_HEADER_SUBJECT, NULL);

  /* TODO: Do quoted printable conversion */

  if (header_value == NULL)
    return 0;

  subject = a_Utils_encode_8bit_subject(header_value, header_charset, 0);

  subject_field = mailimf_subject_new(subject);
  if (subject_field)
    {
      field = mailimf_field_new(MAILIMF_FIELD_SUBJECT,
				NULL, NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, subject_field, NULL,
				NULL, NULL);
      if (field)
	{
	  r = clist_append(fields->list, field);
	  if (r)
	    {
	      A_LOGGING_SYSTEM_ERROR("clist_append() failed");
	      mailimf_field_free(field);
	    }
	}
      else
	{
	  A_LOGGING_SYSTEM_ERROR("mailimf_field_new() failed");
	  mailimf_subject_free(subject_field);
	  r = -1;
	}
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("mailimf_subject_new() failed");
      r = -1;
    }
  return r;
}

/* From: header functions */

int
a_Message_header_set_from(XePMessage *message, 
			  const char *from,
			  const char *charset)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_field *field;
  struct mailimf_from *from_field;
  struct mailimf_mailbox_list *mb_list;
  struct mailimf_mailbox *mb;
  clist *list;
  int cur_token = 0;
  int r = 0;

  A_LOGGING_TRACE;

  a_Message_header_remove(message, XEP_MAIL_HEADER_FROM, NULL);

  if (from == NULL)
    {
      return 0;
    }

  list = clist_new();
  if (list == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new() failed");
      return -1;
    }

  /* Create and encode the list of mailboxes in the From header */
  while ((cur_token < strlen(from)) && (r == 0))
    {
      r = mailimf_mailbox_parse((char *) from, strlen(from),
				&cur_token, &mb);
      if (r == 0)
	{
	  if (mb->display_name)
	    {
	      char *qp;
	      if (a_Utils_has_8bit(mb->display_name, 0))
		{
		  qp = a_Utils_encode_8bit_header(mb->display_name, 
						  charset, 1, 1, 0);
		  if (qp)
		    {
		      free(mb->display_name);
		      mb->display_name = qp;
		    }
		}
	    }
	  r = clist_append(list, mb);
	  if (r)
	    {
	      A_LOGGING_SYSTEM_ERROR("clist_append() failed");
	    }
	}
      else
	{
	  A_LOGGING_ERROR("mailimf_mailbox_parse() failed");
	}
      cur_token++;
    }

  /* Create and add the mailbox_list */
  if (r == 0 && (mb_list = mailimf_mailbox_list_new(list)))
    {
      from_field = mailimf_from_new(mb_list);
      if (from_field)
	{
	  field = mailimf_field_new(MAILIMF_FIELD_FROM,
				    NULL, NULL, NULL, NULL, NULL,
				    NULL, NULL, NULL, NULL, from_field,
				    NULL, NULL, NULL, NULL, NULL,
				    NULL, NULL, NULL, NULL, NULL,
				    NULL, NULL);
	  if (field)
	    {
	      r = clist_append((fields->list), field);
	      if (r)
		{
		  A_LOGGING_SYSTEM_ERROR("clist_append() failed");
		  mailimf_field_free(field);
		}
	    }
	  else
	    {
	      A_LOGGING_SYSTEM_ERROR("mailimf_field_new() failed");
	      mailimf_from_free(from_field);
	      r = -1;
	    }
	}
      else
	{
	  mailimf_mailbox_list_free(mb_list);
	  r = -1;
	}
    }
  else
    {
      clistiter *iter;
      for (iter = clist_begin(list); iter; iter = iter->next)
	{
	  mailimf_mailbox_free((struct mailimf_mailbox *) iter->data);
	}
      clist_free(list);
    }

  return r;
}

int
a_Message_message_get_from(struct mailimf_single_fields *fields,
			   char *dest_charset,
			   char **display_name,
			   char **email_address)
{
  struct mailimf_mailbox *mb_name;
  char *decoded_from = NULL;
  size_t cur_token;
  int r;
  
  A_LOGGING_TRACE;

  if (fields->from == NULL || clist_isempty(fields->from->mb_list->list))
    return -1;

  mb_name = clist_begin(fields->from->mb_list->list)->data;

  if (!mb_name || (!mb_name->display_name && !mb_name->addr_spec))
    return -1;

  if (mb_name->display_name)
    {
      cur_token = 0;
      r = mailmime_encoded_phrase_parse(DEFAULT_SOURCE_CHARSET,
					mb_name->display_name,
					strlen(mb_name->display_name),
					&cur_token,
					dest_charset,
					&decoded_from);
      if (r != MAILIMF_NO_ERROR)
	{
	  *display_name = strdup(mb_name->display_name);
	}
      else
	{
	  *display_name = decoded_from;
	}
    }

  /* Convert even the email address: do not assume that US-ASCII does not need
   * to be converted.
   */
  if (mb_name->addr_spec)
    {
      decoded_from = NULL;
      cur_token = 0;
      r = mailmime_encoded_phrase_parse(DEFAULT_SOURCE_CHARSET,
					mb_name->addr_spec,
					strlen(mb_name->addr_spec),
					&cur_token,
					dest_charset,
					&decoded_from);
      if (r != MAILIMF_NO_ERROR)
	{
	  *email_address = strdup(mb_name->addr_spec);
	}
      else
	{
	  *email_address = decoded_from;
	}
    }      

  return 0;  
}

char *
a_Message_message_formatted_get_from(struct mailimf_single_fields *fields,
				     char *template,
				     char *dest_charset)
{
  XePUtilsFormat format[] =
    {
      {NULL, 'f'},
      {NULL, 'F'},
      {NULL, 0}
    };
  char *value = NULL;
  char *display_name = NULL;
  char *email_address = NULL;
  if (a_Message_message_get_from(fields, 
				 dest_charset,
				 &display_name,
				 &email_address) == 0)
    {
      format[0].string = display_name;
      format[1].string = email_address;
      value = a_Utils_printf(template, format);
      x_free(display_name);
      x_free(email_address);
    }

  return value;
}


const char *
a_Message_get_first_from_addr_spec(XePMessage *message)
{
  struct mailimf_field *field;
  field = a_Message_get_field(message, XEP_MAIL_HEADER_FROM, NULL);

  if (field == NULL || field->field.from == NULL || 
      field->field.from->mb_list->list == NULL   || 
      field->field.from->mb_list->list->first == NULL ||
      field->field.from->mb_list->list->first->data == NULL)
    {
      return NULL;
    }
  else
    {
      return ((struct mailimf_mailbox *) field->field.from->mb_list->list->first->data)->addr_spec;
    }

}

/* To: header functions */

int
a_Message_header_add_to(XePMessage *message, 
			const char *to,
			const char *charset)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_address_list *addr_list;
  int r = 0;
  A_LOGGING_TRACE;

  addr_list = Message_header_get_addr_list(fields, MAILIMF_FIELD_TO);

  r = Message_address_list_add(addr_list, to, charset);
  return r;
}

int
a_Message_message_get_to(struct mailimf_single_fields *fields,
			 char *dest_charset,
			 char **to)
{
  return -1;
}

char *
a_Message_message_formatted_get_to(struct mailimf_single_fields *fields,
				char *template,
				char *dest_charset)
{
  char *result;
  XePUtilsFormat format[] =
    {
      {NULL, 't'},
      {NULL, 'T'},
      {NULL, 0}
    };

  if (fields->to == NULL || fields->to->addr_list == NULL)
    return NULL;

  result = Message_message_formatted_addr_list(fields->to->addr_list->list,
					       template,
					       format,
					       dest_charset);

  return result;
}

/* CC: header functions */

char *
a_Message_message_formatted_get_cc(struct mailimf_single_fields *fields,
				char *template,
				char *dest_charset)
{
  char *result;
  XePUtilsFormat format[] =
    {
      {NULL, 'c'},
      {NULL, 'C'},
      {NULL, 0}
    };

  if (fields->cc == NULL || fields->cc->addr_list == NULL)
    return NULL;

  result = Message_message_formatted_addr_list(fields->cc->addr_list->list,
					       template,
					       format,
					       dest_charset);

  return result;

  
}

int
a_Message_header_add_cc(XePMessage *message,
			const char *cc,
			const char *charset)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_address_list *addr_list;
  int r = 0;

  A_LOGGING_TRACE;

  addr_list = Message_header_get_addr_list(fields, MAILIMF_FIELD_CC);

  r = Message_address_list_add(addr_list, cc, charset);
  return r;
}

/* Bcc: header functions */

int
a_Message_header_add_bcc(XePMessage *message, 
			 const char *bcc, 
			 const char *charset)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_address_list *addr_list;
  int r = 0;

  A_LOGGING_TRACE;

  addr_list = Message_header_get_addr_list(fields, MAILIMF_FIELD_BCC);

  r = Message_address_list_add(addr_list, bcc, charset);
  return r;
}

char *
a_Message_message_formatted_get_bcc(struct mailimf_single_fields *fields,
				 char *template,
				 char *dest_charset)
{

  char *result;
  XePUtilsFormat format[] =
    {
      {NULL, 'b'},
      {NULL, 'B'},
      {NULL, 0}
    };

  if (fields->bcc == NULL || fields->bcc->addr_list == NULL)
    return NULL;

  result = Message_message_formatted_addr_list(fields->bcc->addr_list->list,
					       template,
					       format,
					       dest_charset);

  return result;

}

/* Reply-To: header functions */

char *
a_Message_message_formatted_get_reply_to(struct mailimf_single_fields *fields,
				      char *template,
				      char *dest_charset)
{
  char *result;
  XePUtilsFormat format[] =
    {
      {NULL, 'r'},
      {NULL, 'R'},
      {NULL, 0}
    };

  if (fields->reply_to == NULL)
    return NULL;
  if (fields->reply_to->addr_list == NULL)
    return NULL;

  result = Message_message_formatted_addr_list(fields->reply_to->addr_list->list,
					       template,
					       format,
					       dest_charset);

  return result;
}

int
a_Message_header_add_reply_to(XePMessage *message,
			      const char *reply_to, 
			      const char *charset)
{
  struct mailimf_fields *fields = message->mime->fields;
  struct mailimf_address_list *addr_list;
  int r = 0;

  A_LOGGING_TRACE;

  addr_list = Message_header_get_addr_list(fields, MAILIMF_FIELD_REPLY_TO);

  r = Message_address_list_add(addr_list, reply_to, charset);
  return r;
}

/* Date: header functions */

char *
a_Message_message_formatted_get_local_date(struct mailimf_single_fields *fields,
					   char *template,
					   char *dest_charset)
{
  char orig_date[DATE_BUF_MAX];
  struct tm orig_tm;
  char *result = NULL;

  if (fields->orig_date == NULL)
    return NULL;

  if (fields->orig_date->date_time == NULL)
    return NULL;

  a_Utils_orig_date_to_localtime(fields->orig_date->date_time, &orig_tm);

  orig_date[0] = 0x00;

  if (strftime(orig_date, DATE_BUF_MAX - 1, template, &orig_tm))
    {
      result = strdup(orig_date);
    }

  return result;
}


/* Sender: header functions */

char *
a_Message_message_formatted_get_sender(struct mailimf_single_fields *fields,
				    char *template,
				    char *dest_charset)
{
  return NULL;
}

/* Message-ID: header functions */

char *
a_Message_message_formatted_get_message_id(struct mailimf_single_fields *fields,
					char *template,
					char *dest_charset)
{
  return NULL;
}

/* In-Reply-To: header functions */

char *
a_Message_message_formatted_get_in_reply_to(struct mailimf_single_fields *fields,
					 char *template,
					 char *dest_charset)
{
  return NULL;
}

/* References: header functions */

char *
a_Message_message_formatted_get_references(struct mailimf_single_fields *fields,
					char *template,
					char *dest_charset)
{
  return NULL;
}


/* User Agent: header functions */

static void
Message_set_user_agent(struct mailimf_fields *fields)
{
  const char *user_agent;
  const char *header_name;
  struct mailimf_field *field;

#if 0
  user_agent = a_Prefs_get_user_agent();
#else
#  ifdef LIBETPAN_VERSION
  user_agent = "XetPan/" VERSION " libEtPan!/" LIBETPAN_VERSION;
#  else
  user_agent = "XetPan/" VERSION " libEtPan!/0.0.0";
#  endif
#endif

  header_name = "User-Agent";
  Message_add_optional_header(fields, header_name, user_agent);
}

/* Comments: header functions */

char *
a_Message_message_formatted_get_comments(struct mailimf_single_fields *fields,
					 char *template,
					 char *dest_charset)
{
  return NULL;
}

/* Keywords: header functions */

char *
a_Message_message_formatted_get_keywords(struct mailimf_single_fields *fields,
					 char *template,
					 char *dest_charset)
{
  return NULL;
}

/* Other headers functions */

static int
Message_add_optional_header(struct mailimf_fields *fields,
			    const char *header_name,
			    const char *header_value)
{
  struct mailimf_optional_field *opt_field = NULL;
  struct mailimf_field *field = NULL;
  char *name;
  char *value = NULL;
  int r = 0;

  if (header_name == NULL)
    {
      return -1;
    }

  name = strdup(header_name);
  if (name == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("strdup() failed");
      return -1;
    }

  if (header_value)
    {
      value = strdup(header_value);
      if (value == NULL)
	{
	  A_LOGGING_SYSTEM_ERROR("strdup() failed");
	  x_free(name);
	  return -1;
	}
    }

  opt_field = mailimf_optional_field_new(name, value);
  if (opt_field)
    {
      field = mailimf_field_new(MAILIMF_FIELD_OPTIONAL_FIELD,
				NULL, NULL, NULL, NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, NULL,
				NULL, NULL, NULL, NULL,
				NULL, NULL, opt_field);
      if (field)
	{
	  r = clist_append(fields->list, field);
	}
    }

  if (r)
    {
      
      if (field)
	{
	  mailimf_field_free(field);
	}
      if (opt_field)
	{
	  mailimf_optional_field_free(opt_field);
	}
      x_free(name);
      x_free(value);
    }

  return r;
}

int
a_Message_header_add_other(XePMessage *message,
			   const char *header_name,
			   const char *header_value,
			   const char *header_charset)
{
  int r = 0;
  A_LOGGING_TRACE;

  return r;
}

int
a_Message_message_get_other(struct mailimf_fields *fields,
			    char *other_header, 
			    char *dest_charset,
			    char **value)
{
  clistiter *iter;
  struct mailimf_field *field = NULL;
  char *decoded = NULL;
  int found = 0;
  int cur_token;
  int r = -1;

  if (fields == NULL)
    return r;

  for (iter = clist_begin(fields->list); iter && !found; iter = iter->next)
    {
      field = (struct mailimf_field *) iter->data;
      if (field->type == MAILIMF_FIELD_OPTIONAL_FIELD &&
	  field->field.optional_field)
	{
	  if (field->field.optional_field->name &&
	      strcasecmp(field->field.optional_field->name, other_header) == 0)
	    {
	      found = 1;
	    }
	}
    }
  
  if (found)
    {
      if (field->field.optional_field->value)
	{
	  cur_token = 0;
	  r = 
	    mailmime_encoded_phrase_parse(DEFAULT_SOURCE_CHARSET,
					  field->field.optional_field->value,
					  strlen(field->field.optional_field->value),
					  &cur_token,
					  dest_charset,
					  &decoded);
	  if (r != MAIL_NO_ERROR)
	    {
	      decoded = strdup((field->field.optional_field->value));
	    }
	}
      else
	{
	  decoded = strdup("");
	  r = 0;
	}
    }
  *value = decoded;
  return r;
}

char *
a_Message_message_formatted_get_other(struct mailimf_single_fields *fields,
				      char *template,
				      char *other_header,
				      char *dest_charset)
{
  return NULL;
}

/* Functions to get message data from its folder/mailbox */

int
a_Message_get_bodystructure(XePMessage *message)
{
  int r = -1;
  mailmessage *msg = message->msg;
  struct mailmime *mime = NULL;
  A_LOGGING_TRACE;

  if (msg == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("mailmessage == NULL");
    }
  else
    {
      r = a_Folder_noop(message->folder);
      if (r == MAIL_NO_ERROR)
	{
	  r = mailmessage_get_bodystructure(msg, &mime);

	  if (r != MAIL_NO_ERROR)
	    {
	      A_LOGGING_ERROR(_("Error while fetching message structure: %s"),
			      maildriver_strerror(r));
	      mime = NULL;
	    }
	  else
	    {
	      struct mailimf_fields *fields = NULL;
      	      r = mailmessage_fetch_envelope(msg, &fields);
	      if (r == MAIL_NO_ERROR)
		{
		  mailmessage_resolve_single_fields(msg);
		}
	      else
		{
		  A_LOGGING_ERROR(_("Error while fetching message headers: %s"),
				  maildriver_strerror(r));
		  mailmime_free(mime);
		  mime = NULL;				   
		}
	    }
	}
    }
  
  message->mime = mime;
  return r;
}

static int
Message_fetch_part_decoded(mailmessage *msg, 
			   struct mailmime *mime,
			   struct mailmime_single_fields *fields,
			   char **data,
			   size_t *len)
{
  int r;
  char *raw_part = NULL;
  size_t raw_len;
  char *decoded_part = NULL;
  size_t decoded_len;
  size_t cur_token;
  int encoding;

  A_LOGGING_TRACE;

  r = mailmessage_fetch_section(msg, mime, &raw_part, &raw_len);
  if (r != MAIL_NO_ERROR)
    {
      A_LOGGING_ERROR("Error while fetching section: %s",
		      maildriver_strerror(r));
      return r;
    }

  if (fields->encoding)
    {
      encoding = fields->encoding->type;
    }
  else
    {
      encoding = MAILMIME_MECHANISM_8BIT;
    }
		     
  cur_token = 0;

  r = mailmime_part_parse(raw_part, raw_len, &cur_token, encoding,
			  &decoded_part, &decoded_len);

  if (r == MAILIMF_NO_ERROR)
    {
      *data = decoded_part;
      *len  = decoded_len;
    }
  else
    {
      A_LOGGING_ERROR("Error while decoding message part");
    }

  mailmessage_fetch_result_free(msg, raw_part);

  A_LOGGING_DEBUG("Message_fetch_part_decoded result = %d", r);
  return r;
}

void
a_Message_decode_selected(XePMessage *message)
{
  int r;
  char *decoded_part;
  size_t decoded_len;
  struct mailmime_data *body;
  size_t cur_token = 0;

  x_free(message->data);
  message->data = NULL;
  message->len = 0;

  if (!message->selected_mime || !message->selected_mime->body)
    return;

  body = message->selected_mime->body;
  message->data = NULL;
  message->len  = 0;
  switch (body->type)
    {
    case MAILMIME_DATA_TEXT:
      r = mailmime_part_parse(body->data, body->length, &cur_token,
			      body->encoding, &decoded_part, &decoded_len);
      if (r == MAILIMF_NO_ERROR)
	{
	  message->data = x_alloc(decoded_len, char);
	  message->len  = decoded_len;
	  memcpy(message->data, decoded_part, decoded_len);
	  mailmime_decoded_part_free(decoded_part);
	}
      else
	{
	  A_LOGGING_ERROR("Error while decoding a message part");
	}
      break;
    case MAILMIME_DATA_FILE:
      r = a_Utils_load_file(body->filename, &message->data, &message->len);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Unknown MAILMIME_DATA type");
    }
}

/*! Fetch a message part
 * \param message the structure containing references to the message and
 *        message part to select (message.msg and message.selected_mime)
 *        If the message part is to be decoded, message.dest_charset must
 *        be set.
 *        On success, message.data and message.len are filled with the
 *        part. The data should be freed by mmap_string_unref.
 *        On error, message.data and message.len are set to NULL and 0
 * \return 0 on success
 */
int
a_Message_fetch_part(XePMessage *message)
{
  mailmessage *msg = message->msg;
  struct mailmime *mime = message->selected_mime;
  struct mailmime_single_fields fields;
  char *original_data = NULL;
  size_t original_len = 0;
  char *result_data = NULL;
  size_t result_len = 0;
  int r = -1;

  A_LOGGING_DEBUG("Message msg=%p mime=%p selected_mime=%p destchar=%s",
		  message->msg, message->mime, message->selected_mime,
		  message->dest_charset);

  if (msg == NULL || mime == NULL )
    {
      A_LOGGING_SYSTEM_ERROR("msg or mime NULL");
    }
  else
    {
      memset(&fields, 0, sizeof(struct mailmime_single_fields));
      
      /* TODO: What if mime->mime_fields is NULL? */
      if (mime->mime_fields)
	mailmime_single_fields_init(&fields,
				    mime->mime_fields,
				    mime->content_type);

      
      r = Message_fetch_part_decoded(msg, mime, &fields, 
				     &original_data, &original_len);
      
      if (!r)
	{
	  /* Data to be freed with mailmime_decoded_part_free() 
	   * which is mmap_string_unref()
	   */
	  message->source_charset = (fields.content_charset ? 
				     strdup(fields.content_charset):
				     NULL);						
	  
	  result_data = x_alloc(original_len, char);
	  memcpy(result_data, original_data, original_len);
	  result_len  = original_len;
	  mailmime_decoded_part_free(original_data);
	}
    }
  message->data = result_data;
  message->len  = result_len;

  return r;
}

int
a_Message_fetch(XePMessage *message)
{
  int r;
  char *data;
  size_t len;
  mailmessage *msg = message->msg;

  A_LOGGING_DEBUG("msg = %p message = %p", message->msg, message);

  if (msg == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("mailmessage = NULL");
      return -1;
    }

  r = mailmessage_fetch(msg, &data, &len);
  if (r != MAIL_NO_ERROR)
    {
      const char *err_msg;

      A_LOGGING_ERROR("while fetching message: %s", maildriver_strerror(r));
      
      message->data = NULL;
      message->len = 0;
    }
  else
    {
      message->data = strdup(data);
      message->len = len;
    }

  mailmessage_fetch_result_free(msg, data);

  return r;
}

int
a_Message_flags_change(XePMessage *message)
{
  int changed = 0;

  A_LOGGING_INFO("Setting flags x%0x unsetting x%0x from msg flags x%0x",
		 message->flags_set, 
		 message->flags_unset,
		 message->msg->flags->flags);

  if (!message->msg || !message->msg->flags)
    {
      A_LOGGING_SYSTEM_ERROR("can't set flags: msg or msg->flags are NULL");
      return -1;
    }

  if ((message->msg->flags->flags & message->flags_set) != message->flags_set)
    {
      message->msg->flags->flags |= message->flags_set;
      A_LOGGING_DEBUG("Flag x%0x set", message->flags_unset);

      changed = 1;
    }
  if ((message->msg->flags->flags & message->flags_unset) != 0)
    {

      message->msg->flags->flags &= ~(message->flags_unset);
      A_LOGGING_DEBUG("Flag x%0x unset", message->flags_unset);
      changed = 1;
    }

  if (changed)
    {
      mailmessage_check(message->msg);
      a_Folder_check(message->folder);
      /* This is to do a STATUS on the folder
       * to update message count, although STATUS should not
       * be used on the selected IMAP folder 
       */
      a_Folder_noop(message->folder);
    }
  message->flags_set = 0;
  message->flags_unset = 0;

  return 0;
}

int
a_Message_array_flags_change(carray *messages)
{
  XePMessage *message;
  int changed = 0;
  int need_folder_check = 0;
  int r = 0;
  int i;
  int count = carray_count(messages);

  A_LOGGING_TRACE;

  for (i = 0; i < count; i++)
    {
      message = (XePMessage *) carray_get(messages, i);
      changed = 0;
      printf("Message %d\n", message->msg->index);
      if (!message->msg || !message->msg->flags)
	{
	  /* TODO: improve message error */
	  A_LOGGING_SYSTEM_ERROR("can't set flags: msg or msg->flags are NULL");
	  r = -1;
	  continue;
	}

      if ((message->msg->flags->flags & message->flags_set) != message->flags_set)
	{
	  message->msg->flags->flags |= message->flags_set;
	  A_LOGGING_DEBUG("Flag x%0x set", message->flags_unset);
	  
	  changed = 1;
	}

      if ((message->msg->flags->flags & message->flags_unset) != 0)
	{
	  message->msg->flags->flags &= ~(message->flags_unset);
	  A_LOGGING_DEBUG("Flag x%0x unset", message->flags_unset);
	  changed = 1;
	}
      
      if (changed)
	{
	  printf("Checked\n");
	  mailmessage_check(message->msg);
	  need_folder_check = 1;
	}
      message->flags_set = 0;
      message->flags_unset = 0;
    }
  
  if (need_folder_check)
    {
      printf("Folder check..\n");
      a_Folder_check(message->folder);
      /* This is to do a STATUS on the folder
       * to update message count, although STATUS should not
       * be used on the selected IMAP folder 
       */
      a_Folder_noop(message->folder);
    }
  
  return r;
}



/* Mime parts handling functions */

/*! Find the default (viewable text) MIME part
 *  Recursively walk the message structure and check each of its parts
 *  and returns the first viewable message part.
 *  \param msg the message to search in
 *  \param mime the mime part to check. At entry point, pass the bodystructure
 *  \return the mime part on success, NULL on error
 */
static struct mailmime *
Message_mime_find_default(mailmessage *msg,
			  struct mailmime *mime)
{
  struct mailmime *selected_mime = NULL;
  clistiter *mime_iter;
  struct mailmime *a_mime;
  int score;
  int preferred_score = 0;

  switch (mime->type)
    {
    case MAILMIME_SINGLE:
      if (a_Message_mime_is_text(mime))
	{
	  selected_mime = mime;
	}
      break;
    case MAILMIME_MULTIPLE:
      if (strcmp(mime->content_type->subtype, "alternative") == 0)
	{
	  A_LOGGING_DEBUG("Message mime type is multipart/alternative");
	  for (mime_iter = clist_begin(mime->list);
	       mime_iter;
	       mime_iter = mime_iter->next)
	    {
	      a_mime = (struct mailmime *) mime_iter->data;
	      score = 1;

	      if (a_Message_mime_is_text(a_mime))
		score = 2;
	      
	      if (a_mime->content_type)
		{
		  if (strcasecmp(a_mime->content_type->subtype, "plain") == 0)
		    score = 3;
		}
	      if (score > preferred_score)
		{
		  preferred_score = score;
		  selected_mime = a_mime;
		}
	    }
	}
      else
	{
	  A_LOGGING_DEBUG("Message mime type is multipart/not-alternative");
	  for (mime_iter = clist_begin(mime->list); 
	       mime_iter && !selected_mime;
	       mime_iter = mime_iter->next)
	    {
	      a_mime = (struct mailmime *) mime_iter->data;
	      selected_mime = Message_mime_find_default(msg, a_mime);
	    }
	}
      break;
    case MAILMIME_MESSAGE:
      A_LOGGING_DEBUG("Message mime type is message/*");
      if (mime->msg_mime)
	{
	  selected_mime = Message_mime_find_default(msg, mime->msg_mime);
	}
      else 
	{
	  A_LOGGING_SYSTEM_ERROR("This is a message/* mime part but no "
				 "part actually found...");
	}

      break;
    }

  return selected_mime;
}

/*! Select the default mime part to be displayed
 *  \param message the structure containing the references to the message
 *         and message's body structure (message.msg and message.mime)
 *         On success, message.selected_mime points to the default mime part
 *         On error, message.selected_mime is set to NULL
 *  \return 0 on success
 */
int
a_Message_mime_find_default(XePMessage *message)
{
  mailmessage *msg = message->msg;
  struct mailmime *mime = message->mime;

  A_LOGGING_DEBUG("Message msg=%p mime=%p", message->msg, message->mime);

  if (msg == NULL || mime == NULL)
    return -1;

  message->default_mime = Message_mime_find_default(msg, mime);
  
  if (message->default_mime == NULL)
    {
      A_LOGGING_INFO("No default mime part found");
    }

  message->selected_mime = message->default_mime;
  return (message->default_mime == NULL);
}

int
a_Message_mime_is_text(struct mailmime *mime)
{
  int r = 0;

  if (mime && mime->type == MAILMIME_SINGLE)
    {
      if (mime->content_type && 
	  mime->content_type->type->type == MAILMIME_TYPE_DISCRETE_TYPE)
	{
	  if (mime->content_type->type->discrete_type->type ==
	      MAILMIME_DISCRETE_TYPE_TEXT)
	    r = -1;
	}
      else
	{
	  r = -1;
	}
    }
  else
    r = 0;
	
  return r;
}


/*! Return mime type of the content as an allocated string.
 *  The string must be freeed by the user.
 */
char *
a_Message_mime_get_type(struct mailmime_content *content)
{
  
  char *mime_type = NULL;
  const char *type = "unknown";
  const char *subtype;
  int ok = -1;

  switch (content->type->type)
    {
    case MAILMIME_TYPE_DISCRETE_TYPE:
      switch (content->type->discrete_type->type)
	{
	case MAILMIME_DISCRETE_TYPE_TEXT:
	  type = "text";
	  break;
	case MAILMIME_DISCRETE_TYPE_IMAGE:
	  type = "image";
	  break;
	case MAILMIME_DISCRETE_TYPE_AUDIO:
	  type = "audio";
	  break;
	case MAILMIME_DISCRETE_TYPE_VIDEO:
	  type = "video";
	  break;
	case MAILMIME_DISCRETE_TYPE_APPLICATION:
	  type = "application";
	  break;
	case MAILMIME_DISCRETE_TYPE_EXTENSION:
	  type = content->type->discrete_type->extension;
	  break;
	default:
	  ok = 0;
	}
      break;
    case MAILMIME_TYPE_COMPOSITE_TYPE:
      switch (content->type->composite_type->type)
	{
	case MAILMIME_COMPOSITE_TYPE_MESSAGE:
	  type = "message";
	  break;
	case MAILMIME_COMPOSITE_TYPE_MULTIPART:
	  type = "multipart";
	  break;
	case MAILMIME_COMPOSITE_TYPE_EXTENSION:
	  type = content->type->composite_type->token;
	  break;
	default:
	  ok = 0;
	}
      break;
    default:
      ok = 0;
    }

  if (!ok)
      A_LOGGING_ERROR("Unknown MIME type");

  subtype = (content->subtype ? content->subtype : "");

  mime_type = x_alloc(strlen(type) + 1 + strlen(subtype) + 1, char);

  if (mime_type == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      mime_type[0] = 0x00;
      strcat(mime_type, type);
      strcat(mime_type, "/");
      strcat(mime_type, content->subtype);
    }

  return mime_type;
}

struct mailmime *
a_Message_mime_new_from_file(const char *filename, 
			     const char *mime_type,
			     int mechanism)
{
  struct mailmime *mime = NULL;
  struct mailmime_fields *mime_fields;
  struct mailmime_content *mime_content;
  struct mailmime_disposition *mime_disposition;
  struct mailmime_mechanism *mime_encoding;
  char *magic_mime_type = NULL;
  char *magic_bare_mime_type = NULL;
  char *filename_copy = NULL;
  char *base_name = NULL;
  char *charset = NULL;
  int is_text = 0;
  int r;

  filename_copy = strdup(filename);
  if (filename_copy == NULL)
    {
      A_LOGGING_ERROR("strdup(): %s", strerror(errno));
      return NULL;
    }

  if (mime_type == NULL)
    {
      magic_mime_type = a_Utils_magic_file_mime_type(filename);
    }
  else
    {
      magic_mime_type = strdup(mime_type);
      if (magic_mime_type == NULL)
	{
	  A_LOGGING_ERROR("strdup(): %s", strerror(errno));
	  x_free(filename_copy);
	  return NULL;
	}
    }

  if (strncasecmp(magic_mime_type, "text/", strlen("text/")) ==  0)
    {
      is_text = 1;
      charset = a_Utils_magic_get_charset(magic_mime_type);
    }

  magic_bare_mime_type = a_Utils_magic_get_bare_mime_type(magic_mime_type);
  if (magic_bare_mime_type)
    {  
      if (mechanism == MAILMIME_MECHANISM_ERROR)
	{
	  mechanism = (is_text ? MAILMIME_MECHANISM_QUOTED_PRINTABLE : 
		       MAILMIME_MECHANISM_BASE64);
	}
  
      base_name = a_Utils_basename(filename);
      
      if (base_name)
	{
	  mime_disposition = 
	    mailmime_disposition_new_with_data(MAILMIME_DISPOSITION_TYPE_ATTACHMENT,
					       base_name, NULL, NULL, NULL, 
					       (size_t) -1);
	  if (mime_disposition)
	    {	    
	      mime_encoding = mailmime_mechanism_new(mechanism, NULL);
	      if (mime_encoding)
		{
		  mime_fields = mailmime_fields_new_with_data(mime_encoding, 
							      NULL, NULL,
							      mime_disposition,
							      NULL);
		  if (mime_fields)
		    {
		      mime_content = 
			mailmime_content_new_with_str(magic_bare_mime_type);
		      if (mime_content)
			{
			  mime = mailmime_new_empty(mime_content, mime_fields);
			  if (mime)
			    {
			      r = mailmime_set_body_file(mime, filename_copy);
			      if (r == MAILIMF_NO_ERROR)
				{
				  if (charset)
				    {
				      a_Message_mime_add_parameter(mime,
								   "charset", 
								   charset);
				    }
				}
			      else
				{
				  mailmime_free(mime);
				  free(filename_copy);
				  mime = NULL;
				}
			    }
			  else
			    {
			      mailmime_content_free(mime_content);
			      mailmime_fields_free(mime_fields);
			      free(filename_copy);
			    }
			}
		      else
			{
			  mailmime_fields_free(mime_fields);
			  free(magic_bare_mime_type);
			  free(filename_copy);
			}
		    }
		  else
		    {
		      mailmime_mechanism_free(mime_encoding);
		      mailmime_disposition_free(mime_disposition);
		      free(magic_bare_mime_type);
		      free(filename_copy);
		    }
		}
	      else
		{
		  mailmime_disposition_free(mime_disposition);
		  free(magic_bare_mime_type);
		  free(filename_copy);
		}
	    }
	  else
	    {
	      free(magic_bare_mime_type);
	      free(filename_copy);
	      free(base_name);
	    }
	}
      else
	{
	  free(magic_bare_mime_type);
	  free(filename_copy);
	}
    }
  x_free(charset);

  return mime;


}

struct mailmime *
a_Message_mime_new_from_data(char *data,
			     size_t size,
			     const char *mime_type,
			     int mechanism)
{
  struct mailmime *mime;

  struct mailmime_fields *mime_fields;
  struct mailmime_content *mime_content;
  char *magic_mime_type = NULL;

  if (mime_type == NULL)
    {
      if (data && size)
	{
	  magic_mime_type = a_Utils_magic_data_mime_type(data, size);
	}
    }
  else
    {
      magic_mime_type = strdup(mime_type);
    }

  mime_fields = mailmime_fields_new_encoding(mechanism);

  mime_content = mailmime_content_new_with_str(magic_mime_type);

  mime = mailmime_new_empty(mime_content, mime_fields);

  mailmime_set_body_text(mime, data, size);

  return mime;
}

void
a_Message_mime_set_body_text(struct mailmime *mime, 
			     char *data, 
			     size_t size)
{
  mailmime_set_body_text(mime, data, size);
}

struct mailmime *
a_Message_mime_new_multipart(char *subtype, int mechanism)
{
  struct mailmime *mime;

  struct mailmime_composite_type *composite_type;
  struct mailmime_type *mime_type;
  struct mailmime_content *mime_content;
  struct mailmime_fields *mime_fields;

  mime_fields = mailmime_fields_new_encoding(mechanism);
  
  composite_type = mailmime_composite_type_new(MAILMIME_COMPOSITE_TYPE_MULTIPART, NULL);

  mime_type = mailmime_type_new(MAILMIME_TYPE_COMPOSITE_TYPE, NULL, composite_type);

  mime_content = mailmime_content_new(mime_type, subtype, NULL);

  mime = mailmime_new_empty(mime_content, mime_fields);

  return mime;
}

struct mailmime *
a_Message_mime_new_message(char *subtype, int mechanism)
{
  return mailmime_new_message_data(NULL);
#if 0
  mime_fields = mailmime_fields_new_encoding(mechanism);
  
  composite_type = mailmime_composite_type_new(MAILMIME_COMPOSITE_TYPE_MESSAGE, NULL);

  mime_type = mailmime_type_new(MAILMIME_TYPE_COMPOSITE_TYPE, NULL, composite_type);

  mime_content = mailmime_content_new(mime_type, subtype, NULL);

  mime = mailmime_new_empty(mime_content, mime_fields);

  return mime;
#endif
}

struct mailmime *
a_Message_mime_add_part_to_selected(XePMessage *message, 
				    char *data,
				    size_t size,
				    const char *mime_type,
				    int mechanism)
{
  struct mailmime *new_mime;

  new_mime = a_Message_mime_new_from_data(data, size, mime_type, mechanism);
  mailmime_add_part(message->selected_mime, new_mime);
  return new_mime;
}

int
a_Message_mime_add_submime(struct mailmime *mime, struct mailmime *submime)
{
  int r;

  while (mime && mime->type == MAILMIME_SINGLE)
    {
      mime = mime->parent;
    }

  r = mailmime_smart_add_part(mime, submime);
  if (r == MAILIMF_ERROR_INVAL)
    {
      A_LOGGING_ERROR(_("Error while adding a MIME part: invalid parent MIME"));
    }
  else if (r == MAILIMF_ERROR_MEMORY)
    {
      A_LOGGING_ERROR(_("Error while adding a MIME part: not enough memory"));
    }
  return r;
}

void
a_Message_mime_set_description(struct mailmime *mime, char *description)
{
  struct mailmime_field *afield;
  clistiter *iter;

  /* Modify existent description */
  for (iter = clist_begin(mime->fields->list); iter; iter = iter->next)
    {
      afield = (struct mailmime_field *) iter->data;
      if (afield->type == MAILMIME_FIELD_DESCRIPTION)
	{
	  x_free(afield->description);
	  afield->description = strdup(description);
	  return;
	}
    }
  
  /* Create a description when no one is present */
  afield = mailmime_field_new(MAILMIME_FIELD_DESCRIPTION, 
			      NULL, NULL, NULL, description, 0, NULL, NULL);
  clist_append(mime->fields->list, afield);
}

int
a_Message_mime_add_parameter(struct mailmime *mime, 
			     const char *param_name,
			     const char *param_value)
{
  char *name;
  char *value;
  struct mailmime_parameter *parameter;
  int r = -1;

  if (param_name == NULL || param_value == NULL)
    {
      return 0;
    }
  
  name = strdup(param_name);
  if (name == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("strdup() failed");
      return r;
    }

  value = strdup(param_value);
  if (value == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("strdup() failed");
      free(value);
      return r;
    }

  parameter = mailmime_parameter_new(name, value);
  if (parameter == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("strdup() failed");
      free(name);
      free(value);
      return r;
    }

  if (clist_append(mime->content_type->parameters, parameter))
    {
      A_LOGGING_SYSTEM_ERROR("clist_append() failed");
      mailmime_parameter_free(parameter);
      return r;
    }

  return 0;
}

int
a_Message_mime_remove_parameter(struct mailmime *mime,
			     const char *param_name)
{
  struct mailmime_parameter *parameter;
  clist *parameters;
  clistiter *iter;
  
  if (mime == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("mime is NULL");
      return -1;
    }
  if (mime->content_type == NULL || mime->content_type->parameters == NULL)
    {
      return 0;
    }

  parameters = mime->content_type->parameters;

  iter = clist_begin(parameters);
  while (iter)
    {
      parameter = (struct mailmime_parameter *) iter->data;
      if (param_name == NULL || strcasecmp(parameter->name, param_name) == 0)
	{
	  clist_delete(parameters, iter);
	  mailmime_parameter_free(parameter);
	  iter = clist_begin(parameters);
	}
      else
	{
	  iter = iter->next;
	}
    }
  return 0;
}

int
a_Message_mime_remove_selected(XePMessage *message)
{
  mailmime_remove_part(message->selected_mime);
  mailmime_free(message->selected_mime);
  message->selected_mime = NULL;
  return 0;
}

/* Message saving functions */

int
a_Message_data_save_as(XePMessage *message, const char *filename)
{
  int r = 0;
  int fd;
  FILE *fstr;

  fstr = fopen(filename, "w");
  if (fstr)
    {
      fd = fileno(fstr);
      r = a_Utils_write(fd, message->data, message->len);
      fclose(fstr);
    }
  else
    {
      A_LOGGING_ERROR(_("Error while saving data in %s: %s"),
		      filename,
		      strerror(errno));
      r = -1;
    }
  return r;
}

void
a_Message_write_to_file(XePMessage *message, FILE *file)
{
  int col = 0;
  mailmime_write(file, &col, message->mime);
}

char *
a_Message_write_to_temp_file(XePMessage *message)
{
  FILE *fstr;
  char filename[] = "xetpan.msg.XXXXXX";
  char *filename_copy = NULL;
  int fd;

  fd = mkstemp(filename);
  if (fd != -1 && (fstr = fdopen(fd, "w")))
    {
      a_Message_write_to_file(message, fstr);
      fclose(fstr);
      filename_copy = strdup(filename);
    }
  else
    {
      A_LOGGING_ERROR(_("Error while attempting to save a message: %s"),
		      strerror(errno));
    }
  
  return filename_copy;
}

void
a_Message_append_to_folder(XePMessage *message)
{
  if (message->fd >= 0)
    {
      x_free(message->msg_data);
      message->msg_data = NULL;
      message->msg_size = 0;
      if (a_Utils_read(message->fd, &message->msg_data, &message->msg_size))
	{
	  A_LOGGING_ERROR("Message not appended.");
	  return;
	}
    }
  if (message->msg_data)
    {
      a_Folder_append(message->folder, message->msg_data, message->msg_size);
    }
  else
    {
      A_LOGGING_ERROR("Attempt to append a void message to folder '%s'",
		      a_Folder_get_id(message->folder));
    }
}

carray *
a_Message_array_new_for_flags_change(XePFolder *folder,
				     void **selection, 
				     uint32_t n, 
				     XePMessageFlagsOp flag_op,
				     int flags)
{
  uint32_t i;
  uint32_t m;
  XePMessage *message;
  carray *messages;
  mailmessage *msg;

  messages = carray_new(n);
  carray_set_size(messages, n);
  
  for (m = 0, i = 0; i < n; i++)
    {
      msg = ((struct mail_envelopes_tree *) selection[i])->msg;
      
      if (msg == NULL)
	continue;

      message = a_Message_new();
      message->folder = folder;
      message->msg = msg;
      switch (flag_op)
	{
	  /* TODO: handle non standard flags */
	case XEP_FLAGS_TOGGLE:
	  message->flags_set = (msg->flags->flags & flags ? 
				0 : flags);
	  message->flags_unset = (msg->flags->flags & flags ? 
				  flags : 0);
	  break;
	case XEP_FLAGS_SET:
	  message->flags_set = flags;
	  break;
	case XEP_FLAGS_UNSET:
	  message->flags_unset = flags;
	}
      
      carray_set(messages, m++, message);
    }
  carray_set_size(messages, m);

  return messages;
}

carray *
a_Message_array_new(XePFolder *folder, void **selection, uint32_t n)
{
  uint32_t i;
  uint32_t m;
  XePMessage *message;
  carray *messages;
  mailmessage *msg;

  messages = carray_new(n);
  carray_set_size(messages, n);
  
  for (m = 0, i = 0; i < n; i++)
    {
      msg = ((struct mail_envelopes_tree *) selection[i])->msg;
      
      if (msg == NULL)
	continue;

      message = a_Message_new();
      message->folder = folder;
      message->msg = msg;
      carray_set(messages, m++, message);
    }
  carray_set_size(messages, m);

  return messages;
}

void
a_Message_array_free(carray *messages)
{
  uint32_t i;
  uint32_t count;
  XePMessage *message;

  count = carray_count(messages);
  for (i = 0; i < count; i++)
    {
      message = (XePMessage *) carray_get(messages, i);
      a_Message_free(message);
    }
  carray_free(messages);
}
