/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h> /* unlink() */

#include <mailsmtp.h>

#include "intl.h"

#include "x_alloc.h"
#include "logging.h"
#include "message.h"
#include "utils.h"
#include "folder.h"

#include "mta.h"

static clist *global_mtas;

static void
MTA_send_message_with_command(XePMTA *mta,
			      XePManagerMsgCb sent_cb, 
			      void *user_data,
			      const char *sent_ref);

static void
MTA_send_message_with_smtp_async(XePMTA *mta,
				 XePManagerMsgCb sent_cb, 
				 void *user_data,
				 const char *sent_ref);

static void
MTA_save_sent_file(const char *filename, const char *sent_ref);

static void
MTA_rcpt_append_addr_list(clist *rcpt_list, 
			  struct mailimf_address_list *addr_list);

/* Function definitions */

clist *
a_MTA_list_new(void)
{
  clist *mtas;

  mtas = clist_new();

  if (mtas == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new failed");
    }

  global_mtas = mtas;

  return mtas;
}

void
a_MTA_list_free(clist *mtas)
{
  clistiter *iter;

  for (iter = clist_begin(mtas); iter; iter = iter->next)
    {
      XePMTA *mta = (XePMTA *) iter->data;
      a_MTA_free(mta);
    }
  clist_free(mtas);
}

clist *
a_MTA_list_get(void)
{
  return global_mtas;
}

XePMTA *
a_MTA_new(void)
{
  XePMTA *mta;

  mta = x_alloc(1, XePMTA);

  if (mta == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      mta->id = NULL;
      mta->type = XEP_MTA_INVALID;
      mta->command = NULL;
      mta->server = a_Server_new();
      mta->thread.state = XEP_MANAGER_THREAD_STATE_NONE;
    }
  return mta;
}

void
a_MTA_free(XePMTA *mta)
{
  x_free(mta->id);
  x_free(mta->command);
  a_Server_free(mta->server);
  x_free(mta);
}

void
a_MTA_set_id(XePMTA *mta, const char *id)
{
  x_free(mta->id);

  if (id)
    mta->id = strdup(id);
  else
    mta->id = NULL;
}

const char *
a_MTA_get_id(XePMTA *mta)
{
  return mta->id;
}

void
a_MTA_set_type(XePMTA *mta, XePMTAType type)
{
  mta->type = type;
}

XePMTAType
a_MTA_get_type(XePMTA *mta)
{
  return mta->type;
}

void
a_MTA_set_command(XePMTA *mta, const char *command)
{
  x_free(mta->command);
  if (command)
    mta->command = strdup(command);
  else
    mta->command = NULL;
}

const char *
a_MTA_get_command(XePMTA *mta)
{
  return mta->command;
}

void
a_MTA_set_server(XePMTA *mta, XePServer *server)
{
  a_Server_free(mta->server);
  mta->server = server;
}

XePServer *
a_MTA_get_server(XePMTA *mta)
{
  return mta->server;
}

XePMTA *
a_MTA_lookup_by_ref(const char *ref)
{
  XePMTA *mta;
  clistiter *iter;
  clist *mtas = global_mtas;

  if (mtas == NULL)
    {
      A_LOGGING_ERROR(_("There are no MTA registered."));
      return NULL;
    }

  for (iter = clist_begin(mtas); iter; iter = iter->next)
    {
      mta = (XePMTA *) iter->data;
      if ((mta->id && ref && strcmp(mta->id, ref) == 0) ||
	  (ref == NULL) )
	{
	  break;
	}
    }

  return (iter ? mta : NULL);
}

void
a_MTA_send_message(XePMTA *mta, XePMessage *message,
		   XePManagerMsgCb sent_cb, void *user_data,
		   const char *sent_ref)
{
  A_LOGGING_TRACE;

  if (message->fd >= 0)
    {
      A_LOGGING_SYSTEM_ERROR("Warning: message file already opened.");
    }
  x_free(message->filename);
  message->filename = a_Message_write_to_temp_file(message);
  a_Message_file_open(message);

  mta->message = message;

  switch (mta->type)
    {
    case XEP_MTA_COMMAND:
      MTA_send_message_with_command(mta, sent_cb, user_data, sent_ref);
      break;
    case XEP_MTA_SMTP:
      MTA_send_message_with_smtp_async(mta, sent_cb, user_data, sent_ref);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Unknown MTA type");
      sent_cb(user_data, (void *) -1);
    }
}

void
a_MTA_send_message_list(XePMTA *mta, clist *message_list, 
			XePManagerMsgCb sent_cb, void *user_data)
{
  A_LOGGING_ERROR("Not implemented yet");
}

static void
MTA_send_message_with_command(XePMTA *mta,
			      XePManagerMsgCb sent_cb, void *user_data,
			      const char *sent_ref)
{
  const char *filename = mta->message->filename;
  char *cmd;
  int fd_mon = 0;
  int r = -1;

  XePUtilsFormat format[] = {{NULL, 's'}, {NULL, 0}};

  A_LOGGING_TRACE;

  if (mta->command == NULL)
    {
      A_LOGGING_ERROR(_("Missing command for MTA '%s'"), mta->id);
    }  
  else 
    {
      if (filename)
	{
	  if (sent_ref)
	    {
	      MTA_save_sent_file(filename, sent_ref);
	    }

	  format[0].string = filename;
	  
	  cmd = a_Utils_printf(mta->command, format);
	  
	  if (cmd[0] == '|')
	    {
	      FILE *fstr;
	      int fd_in;
	      
	      fstr = fopen(filename, "r");
	      if (fstr && (fd_in = fileno(fstr)) != -1)
		{
		  /*! TODO: the result of the command should be monitored */
		  a_Utils_exec_cmd(cmd + 1, fd_in, fileno(stdout), fileno(stderr),
				   &fd_mon, NULL, 0);
		  r = fd_mon;
		}
	      fclose(fstr);
	    }
	  else
	    {
	      /*! TODO: the result of the command should be monitored */
	      a_Utils_exec_cmd(cmd, fileno(stdin), fileno(stdout), fileno(stderr),
			       &fd_mon, NULL, 0);
	      r = fd_mon;
	    }
	  
	  unlink(filename);
	  
	  free(cmd);

	  mta->message = NULL;
	}
    }

  sent_cb(user_data, (void *) r);
}

static void
MTA_send_message_with_smtp_async(XePMTA *mta,
				 XePManagerMsgCb sent_cb, void *user_data,
				 const char *sent_ref)
{
  XePManagerMsg *msg;
  XePMessage *save_message;

  save_message = a_Message_new_from_file_noparse(mta->message->filename);

  if (sent_ref)
    {
      MTA_save_sent_file(mta->message->filename, sent_ref);
    }

  msg = a_Manager_msg_new();
  a_Manager_msg_set_type(msg, XEP_MANAGER_SMTP_SEND);
  a_Manager_msg_set_source_target(msg, NULL, &mta->thread);
  a_Manager_msg_set_args(msg, mta, &mta->result);
  a_Manager_msg_add_callback(msg, sent_cb, user_data);
  a_Manager_msg_submit(msg);
}

int
a_MTA_smtp_send_message(XePMTA *mta)
{
  struct mailimf_field *field;
  mailsmtp *ep_smtp;
  char *msg_data = NULL;
  size_t msg_size;
  const char *filename = mta->message->filename;
  const char *from = NULL;
  clist *rcpt_list;
  int r = 0;

  if (filename == NULL)
    {
      return -1;
    }

  if (a_Utils_load_file(filename, &msg_data, &msg_size))
    {
      unlink(filename);
      return -1;
    }

  /* Create recipients list */

  /* FROM: */
  from = a_Message_get_first_from_addr_spec(mta->message);
  if (from == NULL)
    {
      A_LOGGING_ERROR(_("No From: field found. Message not sent."));
      x_free(msg_data);
      unlink(filename);
      return -1;
    }
  
  rcpt_list = esmtp_address_list_new();

  /* TO */
  field = a_Message_get_field(mta->message, XEP_MAIL_HEADER_TO, NULL);
  if (field && field->field.to && field->field.to->addr_list)
    {
      MTA_rcpt_append_addr_list(rcpt_list, field->field.to->addr_list);
    }
  /* CC */
  field = a_Message_get_field(mta->message, XEP_MAIL_HEADER_CC, NULL);
  if (field && field->field.cc && field->field.cc->addr_list)
    {
      MTA_rcpt_append_addr_list(rcpt_list, field->field.cc->addr_list);
    }
  /* BCC */
  field = a_Message_get_field(mta->message, XEP_MAIL_HEADER_BCC, NULL);
  if (field && field->field.bcc && field->field.bcc->addr_list)
    {
      MTA_rcpt_append_addr_list(rcpt_list, field->field.bcc->addr_list);
    }
  if (clist_count(rcpt_list) == 0)
    {
      A_LOGGING_ERROR(_("No recipients found. Message not sent."));
      x_free(msg_data);
      unlink(filename);
      esmtp_address_list_free(rcpt_list);
      return -1;
    }

  ep_smtp = mailsmtp_new(0, NULL);
  if (ep_smtp)
    {
      /* Open connection */

      r = mailsmtp_socket_connect(ep_smtp,
				  mta->server->host, mta->server->port);
      if (r == MAILSMTP_NO_ERROR)
	{
	  r = mailsmtp_init(ep_smtp);
	  if (r == MAILSMTP_NO_ERROR)
	    {
	      r = mailesmtp_send(ep_smtp, from, 0, NULL, rcpt_list, 
				 msg_data, msg_size);
	      if (r != MAILSMTP_NO_ERROR)
		{
		  A_LOGGING_ERROR(_("while sending message through SMTP %s "
				    "(host %s): %s"),
				  mta->id, mta->server->host,
				  mailsmtp_strerror(r));
		}
	    }
	  else
	    {
	      A_LOGGING_ERROR(_("while greeting SMTP %s (host %s): %s"),
			      mta->id, mta->server->host,
			      mailsmtp_strerror(r));
	      
	    }
	  mailsmtp_quit(ep_smtp);
	}
      else
	{
	  A_LOGGING_ERROR(_("unable to connect to SMTP %s (host %s): %s"),
			  mta->id, mta->server->host, mailsmtp_strerror(r));

	}
      mailsmtp_free(ep_smtp);
      
    }
  else
    {
      A_LOGGING_ERROR("mailsmtp_new() failed");
    }

  x_free(msg_data);
  unlink(filename);

  esmtp_address_list_free(rcpt_list);
  mta->message = NULL;

  return (r == 0 ? 0 : -1);
}

static void
MTA_save_sent_file(const char *filename, const char *sent_ref)
{
  XePFolder *sent_folder;
  if (sent_ref)
    {
      sent_folder = a_Folder_get_from_ref(sent_ref);
      if (sent_folder)
	{
	  a_Folder_append_file_async(sent_folder,
				     filename,
				     NULL, NULL);
	}
      else
	{
	  A_LOGGING_ERROR("Folder '%s' unknown. Message not saved.", sent_ref);
	}
    }
}

static void
MTA_rcpt_append_addr_list(clist *rcpt_list, struct mailimf_address_list *addr_list)
{
  clistiter *iter;
  clistiter *mbox_iter;
  struct mailimf_address *addr;
  struct mailimf_mailbox *mbox;
  for (iter = clist_begin(addr_list->list); iter; iter = iter->next)
    {
      addr = (struct mailimf_address *) iter->data;
      if (addr->type == MAILIMF_ADDRESS_MAILBOX)
	{
	  esmtp_address_list_add(rcpt_list, addr->mailbox->addr_spec, 0, NULL);
	}
      else if (addr->type == MAILIMF_ADDRESS_GROUP)
	{
	  for (mbox_iter = clist_begin(addr->group->mb_list->list);
	       mbox_iter;
	       mbox_iter = mbox_iter->next)
	    {
	      mbox = (struct mailimf_mailbox *) iter->data;
	      esmtp_address_list_add(rcpt_list, mbox->addr_spec, 0, NULL);
	    }
	}
    }
}
