/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <clist.h>
#include <libetpan.h> /* mailmessage */
#include <stdint.h> /* uint32_t, could be in inttypes.h or sys/types.h, google AC_NEED_STDINT_H */

#include "intl.h"

#include "logging.h"
#include "x_alloc.h"

#include "lxml.h"
#include "message.h" /* message->msg */
#include "processing.h"

typedef enum
  {
    XEP_BOOLEAN = 0,
    XEP_EXPRESSION
  } XePConditionNodeType;

typedef enum
  {
    XEP_ATTRIBUTE_NONE = 0,
    XEP_ATTRIBUTE_CC,
    XEP_ATTRIBUTE_DATE_H,
    XEP_ATTRIBUTE_DATE_D,
    XEP_ATTRIBUTE_FOLDER,
    XEP_ATTRIBUTE_FROM,
    XEP_ATTRIBUTE_SIZE,
    XEP_ATTRIBUTE_SUBJECT,
    XEP_ATTRIBUTE_UID,
    XEP_ATTRIBUTE_TO,
    XEP_ATTRIBUTE_BODY,
    XEP_ATTRIBUTE_AGE,
    XEP_ATTRIBUTE_SCORE,
    XEP_ATTRIBUTE_FLAGS,
    XEP_ATTRIBUTE_SHELL,         /**< return 0 means TRUE, otherwise FALSE */
    XEP_ATTRIBUTE_PYTHON,        /**< Python user function with a specific interface */
    XEP_ATTRIBUTE_FILTER,        /**< For future usage and extension. C-function */
    XEP_ATTRIBUTE_HEADER         /**< User defined header, must be last and needs special treatment */
  } XePExpressionAttributeType;

typedef enum
  {
    XEP_EXPRESSION_TEST_NONE = 0,
    XEP_EXPRESSION_TEST_IS,
    XEP_EXPRESSION_TEST_IS_CASESENSITIVE,
    XEP_EXPRESSION_TEST_HAS,
    XEP_EXPRESSION_TEST_HAS_CASESENSITIVE,
    XEP_EXPRESSION_TEST_REGEXP,
    XEP_EXPRESSION_TEST_REGEXP_CASESENSITIVE,
    XEP_EXPRESSION_TEST_EQ,
    XEP_EXPRESSION_TEST_NE,
    XEP_EXPRESSION_TEST_GT,
    XEP_EXPRESSION_TEST_GE,
    XEP_EXPRESSION_TEST_LT,
    XEP_EXPRESSION_TEST_LE,
    XEP_EXPRESSION_TEST_SET,
    XEP_EXPRESSION_TEST_UNSET
  } XePExpressionTestType;

typedef enum
  {
    XEP_EXPRESSION_VALUE_NONE = 0,
    XEP_EXPRESSION_VALUE_STRING, /**< Quoted string. Escape character "\", 
				  *   dynamic fields: %ATTRIBUTE%, % escaped with \ */
    XEP_EXPRESSION_VALUE_SIZE,   /**< [0-9]+([kMG]?b?) */
    XEP_EXPRESSION_VALUE_NUMBER, /**< [+-][0-9]+ 0x[0-9] */
    XEP_EXPRESSION_VALUE_UID,    /**< [0-9]+ */
    XEP_EXPRESSION_VALUE_DATE,   /**< YYYY-MM-DD MM-DD DD [TZ] */
    XEP_EXPRESSION_VALUE_AGE,    /**< [0-9][ymdh']? : years/months/days/hours/minutes */
    XEP_EXPRESSION_VALUE_FLAGS,  /**< ([+-]?([nud]|[^+- ]*) : e.g.: +n+u-hide */
    XEP_EXPRESSION_VALUE_SHELL,  /**< Shell script: quoted string */
    XEP_EXPRESSION_VALUE_PYTHON, /**< Python script: MODULE FUNCTION STRING */
    XEP_EXPRESSION_VALUE_FILTER  /**< Internal filter: FUNCTION_NAME STRING */
  } XePExpressionValueType;

typedef enum
  {
    XEP_BOOLEAN_NONE = 0,
    XEP_BOOLEAN_AND,
    XEP_BOOLEAN_OR,
    XEP_BOOLEAN_XOR
  } XePBooleanType;

typedef struct XePExpressionValue_s XePExpressionValue;
struct XePExpressionValue_s
{
  XePExpressionValueType type;
  clist *values;
  union
  {
    char *string;
    size_t size;
    long int number; /* mainly used for date? */
    uint32_t uid; 
    struct {
      int flags_on;
      int flags_off;
    } flags;
    struct tm date;
  } value;
};

struct XePProcessingConditionTree_s
{
  XePConditionNodeType node_type;
  int negation;
  XePProcessingConditionTree *parent;
  char *attr_header;
  union
  {
    XePBooleanType boolean;
    XePExpressionTestType test;
  } type;
  union
  {
    XePProcessingConditionTree *node;
    XePExpressionAttributeType type;
  } op1;
  union
  {
    XePProcessingConditionTree *node;
    XePExpressionValue value;
  } op2;
};

typedef struct XePExpressionTest_s XePExpressionTest;
struct XePExpressionTest_s
{
  const char *token;
  XePExpressionTestType type;
};

typedef struct XePExpressionToken_s XePExpressionToken;
struct XePExpressionToken_s
{
  const char *token;
  int type;
  XePExpressionValueType value_type;
};

static int
Processing_conditions_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_actions_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_rules_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_condition_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_action_cb(cntree *atree, void **data, void **data_iter, int op);

static int
Processing_rule_cb(cntree *atree, void **data, void **data_iter, int op);

static void *
Processing_get_from_ref(chash *hash, char *ref);

static int
Processing_update_rule_links(XePProcessingRule *rule);

static XePProcessingConditionTree *
Processing_condition_node_new(XePConditionNodeType node_type,
			      XePBooleanType boolean_type,
			      XePExpressionAttributeType attr_type, 
			      char *attr_header, 
			      XePExpressionTestType test_type,
			      XePExpressionValue *value);

static void
Processing_condition_node_free(XePProcessingConditionTree *node);

static XePProcessingConditionTree *
Processing_condition_parse_atom(const char *expr, size_t *index);

static XePProcessingConditionTree *
Processing_condition_parse_expression(const char *expr, size_t *index, int *depth);

static XePProcessingConditionTree *
Processing_condition_compile(const char *expr, size_t *index, int *depth);

/* Static global variables */

static XePProcessing *global_processing;


XePProcessing *
a_Processing_new(void)
{
  chash *rules;
  chash *conditions;
  chash *actions;
  XePProcessing *processing;

  rules = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (rules == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new returned NULL");
      goto err_rules;
    }

  conditions = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (conditions == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new returned NULL");
      goto err_conditions;
    }

  actions = chash_new(CHASH_DEFAULTSIZE, CHASH_COPYNONE);

  if (actions == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("chash_new returned NULL");
      goto err_actions;
    }

  processing = x_alloc(1, XePProcessing);

  if (processing == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
      goto err_processing;
    }

  processing->rules      = rules;
  processing->conditions = conditions;
  processing->actions    = actions;

  global_processing = processing;

  return processing;

 err_processing:
  chash_free(actions);
 err_actions:
  chash_free(conditions);
 err_conditions:
  chash_free(rules);
 err_rules:
  
  global_processing = NULL;
  return NULL;
}

void
a_Processing_free(void)
{
  chashiter  *hiter;
  chashdatum hiter_datum;
  chash *hash;
  XePProcessing *processing = global_processing;
  
  if (processing == NULL)
    return;

  if ((hash = processing->rules))
    {
      for (hiter = chash_begin(hash); hiter; hiter = chash_next(hash, hiter))
	{
	  chash_value(hiter, &hiter_datum);
	  if (hiter_datum.data)
	    a_Processing_rule_free((XePProcessingRule *)hiter_datum.data);
	}
      chash_free(hash);
    }

  if ((hash = processing->conditions))
    {
      for (hiter = chash_begin(hash); hiter; hiter = chash_next(hash, hiter))
	{
	  chash_value(hiter, &hiter_datum);
	  if(hiter_datum.data)
	    a_Processing_condition_free((XePProcessingCondition *)
					hiter_datum.data);
	}
      chash_free(hash);
    }

  if ((hash = processing->actions))
    {
      for (hiter = chash_begin(hash); hiter; hiter = chash_next(hash, hiter))
	{
	  chash_value(hiter, &hiter_datum);
	  if (hiter_datum.data)
	    a_Processing_action_free((XePProcessingAction *)hiter_datum.data);
	}
      chash_free(hash);
    }
  x_free(processing);
  global_processing = NULL;
}

XePProcessingCondition *
a_Processing_condition_new(void)
{
  XePProcessingCondition *condition;

  condition = x_alloc(1, XePProcessingCondition);

  if (condition)
    {
      condition->id = NULL;
      condition->expression = NULL;
      condition->tree = NULL;
      condition->enabled = 1;
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }

  return condition;
}

void
a_Processing_condition_free(XePProcessingCondition *condition)
{
  if (condition == NULL)
    return;

  A_LOGGING_DEBUG("freeing condition '%s'", condition->id);

  x_free(condition->id);
  x_free(condition->expression);

  x_free(condition);
}

int
a_Processing_condition_add_new(const char *id, const char *expression)
{
  XePProcessingCondition *condition;
  condition = a_Processing_condition_new();
  condition->id = strdup(id);
  condition->expression = strdup(expression);
  return a_Processing_condition_add(condition);
}

int
a_Processing_condition_add(XePProcessingCondition *condition)
{
  chashdatum key;
  chashdatum data;

  if (!global_processing || !condition)
    {
      A_LOGGING_SYSTEM_ERROR("processing or condition NULL.");
       return -1;
    }

  if (a_Processing_condition_compile(condition) != 0)
    {
      return -1;
    }

  A_LOGGING_DEBUG("adding condition '%s'", condition->id);
  
  key.data = condition->id;
  key.len  = strlen(condition->id);

  if (chash_get(global_processing->conditions, &key, &data) == 0)
    {
      A_LOGGING_ERROR("A condition with id='%s' is already here. "
		      "New condition ignored.", condition->id);
      goto err;
    }

  data.data = (char *) condition;
  data.len  = sizeof(condition);

  if (chash_set(global_processing->conditions, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set error. Condition '%s' ignored.",
			     condition->id);
      goto err;
    }

  return 0;

 err:
  a_Processing_condition_free(condition);
  return -1;
}

static void *
Processing_condition_begin_iter(void)
{
  return chash_begin(global_processing->conditions);
}

static void *
Processing_condition_next_iter(void *iter)
{
  return chash_next(global_processing->conditions, (chashiter *) iter);
}

static XePProcessingCondition *
Processing_condition_get_from_iter(void *iter)
{
  chashdatum result;
  chash_value(iter, &result);
  return (XePProcessingCondition *) result.data;
}

XePProcessingAction *
a_Processing_action_new(void)
{
  XePProcessingAction *action;

  action = x_alloc(1, XePProcessingAction);

  if (action)
    {
      action->id = NULL;
      action->command = NULL;
      action->enabled = 1;
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("calloc: %s", strerror(errno));
    }

  return action;
}

void
a_Processing_action_free(XePProcessingAction *action)
{
  if (action == NULL)
    return;

  A_LOGGING_DEBUG("freeing action '%s'", action->id);

  free(action->id);

  free(action->command);

  free(action);
}

int
a_Processing_action_add_new(const char *id, const char *command)
{
  XePProcessingAction *action;
  action = a_Processing_action_new();
  action->id = strdup(id);
  action->command = strdup(command);
  return a_Processing_action_add(action);
}

int
a_Processing_action_add(XePProcessingAction *action)
{
  chashdatum key;
  chashdatum data;

  if (!global_processing || !action)
    {
      A_LOGGING_SYSTEM_ERROR("processing or action NULL.");
       return -1;
    }

  A_LOGGING_DEBUG("adding action %s", action->id);

  key.data = action->id;
  key.len  = strlen(action->id);

  if (chash_get(global_processing->actions, &key, &data) == 0)
    {
      A_LOGGING_ERROR("An action with id='%s' is already here. "
		      "New action ignored.", action->id);
      goto err;
    }

  data.data = (char *) action;
  data.len  = sizeof(action);

  if (chash_set(global_processing->actions, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set error. Action '%s' ignored.",
		      action->id);
      goto err;
    }

  return 0;

 err:
  a_Processing_action_free(action);
  return -1;

}

static void *
Processing_action_begin_iter(void)
{
  return chash_begin(global_processing->actions);
}

static void *
Processing_action_next_iter(void *iter)
{
  return chash_next(global_processing->actions, (chashiter *) iter);
}

static XePProcessingAction *
Processing_action_get_from_iter(void *iter)
{
  chashdatum result;
  chash_value(iter, &result);
  return (XePProcessingAction *) result.data;
}

XePProcessingRule *
a_Processing_rule_new(void)
{
  XePProcessingRule *rule;

  rule = x_alloc(1, XePProcessingRule);

  if (rule)
    {
      rule->id = NULL;
      rule->condition_ref = NULL;
      rule->condition = NULL;
      rule->action_ref = NULL;
      rule->action = NULL;
      rule->enabled = 1;
    }
  else
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }

  return rule;
}

void
a_Processing_rule_free(XePProcessingRule *rule)
{
  if (rule == NULL)
    return;
  
  A_LOGGING_DEBUG("freeing rule '%s'", rule->id);

  x_free(rule->id);
  x_free(rule->condition_ref);
  x_free(rule->action_ref);  

  x_free(rule);
}

XePProcessingRef *
a_Processing_ref_new(void)
{
  XePProcessingRef *ref;

  ref = x_alloc(1, XePProcessingRef);
 
  if (ref == NULL)
    A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
  else
    {
      ref->ref = NULL;
      ref->enabled = 1;
      ref->periodicity = 0;
    }
  
  return ref;
}

XePProcessingRef *
a_Processing_ref_new_copy(XePProcessingRef *ref)
{
  XePProcessingRef *ref_copy;
  
  ref_copy = a_Processing_ref_new();

  if (ref_copy)
    {
      ref_copy->ref         = X_STRDUP(ref->ref);
      ref_copy->enabled     = ref->enabled;
      ref_copy->periodicity = ref->periodicity;
    }

  return ref_copy;  
}

void
a_Processing_ref_free(XePProcessingRef *ref)
{
  if (ref == NULL)
    return;
  
  x_free(ref->ref);
  x_free(ref);
}

clist *
a_Processing_ref_list_new(void)
{
  return clist_new();
}

void
a_Processing_ref_list_free(clist *refs)
{
  clistiter *iter;

  if (refs == NULL)
    return;

  for (iter = clist_begin(refs); iter; iter = iter->next)
    a_Processing_ref_free((XePProcessingRef *)iter->data);

  clist_free(refs);
}

int
a_Processing_rule_add_new(const char *id,
			  const char *condition_ref,
			  const char *action_ref)
{
  XePProcessingRule *rule;
  int r;

  rule = a_Processing_rule_new();
  rule->id = strdup(id);
  rule->condition_ref = strdup(condition_ref);
  rule->action_ref = strdup(action_ref);
  rule->enabled = 1;

  r = Processing_update_rule_links(rule);
  if (r == 0)
    r = a_Processing_rule_add(rule);
  else
    a_Processing_rule_free(rule);

  return r;  
}

int
a_Processing_rule_add(XePProcessingRule *rule)
{
  chashdatum key;
  chashdatum data;

  if (!global_processing || !rule)
    {
      A_LOGGING_SYSTEM_ERROR("processing or rule NULL.");
       return -1;
    }

  A_LOGGING_DEBUG("adding rule '%s'", rule->id);

  key.data = rule->id;
  key.len  = strlen(rule->id);

  if (chash_get(global_processing->rules, &key, &data) == 0)
    {
      A_LOGGING_ERROR("A rule with id='%s' is already here. "
		      "New rule ignored.", rule->id);
      goto err;
    }

  data.data = (char *) rule;
  data.len  = sizeof(rule);

  if (chash_set(global_processing->rules, &key, &data, NULL))
    {
      A_LOGGING_SYSTEM_ERROR("chash_set error. Rule '%s' ignored.",
		      rule->id);
      goto err;
    }

  return 0;

 err:
  a_Processing_rule_free(rule);
  return -1;
}

static void *
Processing_rule_begin_iter(void)
{
  return chash_begin(global_processing->rules);
}

static void *
Processing_rule_next_iter(void *iter)
{
  return chash_next(global_processing->rules, (chashiter *) iter);
}

static XePProcessingRule *
Processing_rule_get_from_iter(void *iter)
{
  chashdatum result;
  chash_value(iter, &result);
  return (XePProcessingRule *) result.data;
}

int
a_Processing_create(cntree *ntree, XePProcessing **processing, int op)
{
  int r;
  LXMLelement *element;
  LXMLelement_callbacks element_cb[] = 
    {
      {"conditions", Processing_conditions_cb, NULL},
      {"actions",    Processing_actions_cb,    NULL},
      {"rules",      Processing_rules_cb,      NULL},
      {NULL, NULL, NULL}
    };
  LXMLelement_callbacks *a_element_cb;

  element = lxml_tree_get_element(ntree);

  for (a_element_cb = element_cb; 
       a_element_cb && a_element_cb->element_name;
       a_element_cb++)
    a_element_cb->data = (void **) global_processing;

  A_LOGGING_DEBUG("Parsing <processing>...");
  r = lxml_tree_convert(ntree, element_cb, op);
  A_LOGGING_DEBUG("<processing> parsed");

  if (op == LXML_Read && r == 0)
    {
      if (a_Processing_update_rules_links())
	{
	  A_LOGGING_ERROR("Processing data discarded");
	  a_Processing_free();
	  a_Processing_new();
	  /* TODO: clean up the switch to a global_processing */
	  *processing = global_processing;
	  return -1;
	}
      if (a_Processing_rules_compile_all())
	{
	  A_LOGGING_ERROR("Some processing rules could not be compiled");
	}
    }
  *processing = global_processing;
  return r;
}

static int
Processing_conditions_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing **processing = (XePProcessing **) data;
  LXMLelement_callbacks element_cb[] = 
    {
      {"condition", Processing_condition_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <condition>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<condition> parsed");

  return r;
}

static int
Processing_actions_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing **processing = (XePProcessing **) data;
  LXMLelement_callbacks element_cb[] = 
    {
      {"action", Processing_action_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <action>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<action> parsed");

  return r;
}

static int
Processing_rules_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessing **processing = (XePProcessing **) data;
  LXMLelement_callbacks element_cb[] = 
    {
      {"rule", Processing_rule_cb, NULL},
      {NULL, NULL, NULL}
    };

  element = lxml_tree_get_element(atree);

  element_cb[0].data = (void **) processing;

  A_LOGGING_DEBUG("Parsing <rule>...");
  r = lxml_tree_convert(atree, element_cb, op);
  A_LOGGING_DEBUG("<rule> parsed");

  return r;
}

static int
Processing_condition_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessingCondition *condition;
  char *id = NULL;
  char *expression = NULL;
  LXMLstore store[] = 
    {
      {"id",         LXML_String, NULL, LXML_Unique},
      {"expression", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };

  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) && 
	  (NULL == (*data_iter = Processing_condition_begin_iter())))
	return 0;

      condition = Processing_condition_get_from_iter(*data_iter);

      *data_iter = Processing_condition_next_iter(*data_iter);

      id = condition->id;
      expression = condition->expression;
    }

  store[0].dest = (void **) &id;
  store[1].dest = (void **) &expression;

  r= lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Condition syntax error. Ignored.");
	  goto err;
	}

      if (!(id && expression))
	{
	  if (id)
	    A_LOGGING_ERROR("Condition '%s' does not have all attributes. "
			    "Ignored.", id);
	  else
	    A_LOGGING_ERROR("Condition is missing the id attribute.");
	  return -1;
	}

      condition = a_Processing_condition_new();

      if (condition == NULL)
	goto err;
      
      condition->id = id;
      condition->expression = expression;
      
      if (a_Processing_condition_add(condition))
	{
	  a_Processing_condition_free(condition);
	  return -1;
	}
    }

  return 0;

 err:
  if (id)
    free(id);
  if (expression)
    free(expression);
  return -1;
}

static int
Processing_action_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessingAction *action;
  char *id = NULL;
  char *command = NULL;
  LXMLstore store[] = 
    {
      {"id",         LXML_String, NULL, LXML_Unique},
      {"command", LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  
  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) &&
	  (NULL == (*data_iter = Processing_action_begin_iter())))
	return 0;
	  
      action = Processing_action_get_from_iter(*data_iter);

      *data_iter = Processing_action_next_iter(*data_iter);

      id = action->id;
      command = action->command;
    }

  store[0].dest = (void **) &id;
  store[1].dest = (void **) &command;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Action syntax error. Ignored.");
	  goto err;
	}
      
      if (!(id && command))
	{
	  if (id)
	    A_LOGGING_ERROR("Action '%s' does not have all attributes."
			    " Ignored.", id);
	  else
	    A_LOGGING_ERROR("Action is missing the id attribute.");
	  return -1;
	}

      action = a_Processing_action_new();
      if (action == NULL)
	goto err;
      
      action->id = id;
      action->command = command;
      
      if (a_Processing_action_add(action))
	{
	  a_Processing_action_free(action);
	  return -1;
	}
    }

  return r;

 err:
  x_free(id);
  x_free(command);
  return -1;
}

static int
Processing_rule_cb(cntree *atree, void **data, void **data_iter, int op)
{
  int r;
  LXMLelement *element;
  XePProcessingRule *rule;
  char *id = NULL;
  char *condition_ref = NULL;
  char *action_ref = NULL;
  LXMLstore store[] = 
    {
      {"id",            LXML_String, NULL, LXML_Unique},
      {"condition_ref", LXML_String, NULL, LXML_Unique},
      {"action_ref",    LXML_String, NULL, LXML_Unique},
      {NULL, 0, NULL, 0}
    };
  
  element = lxml_tree_get_element(atree);

  if (op == LXML_Write)
    {
      if ((NULL == *data_iter) &&
	  (NULL == (*data_iter = Processing_rule_begin_iter())))
	return 0;

      rule = Processing_rule_get_from_iter(*data_iter);

      *data_iter = Processing_rule_next_iter(*data_iter);

      id = rule->id;
      condition_ref = (rule->condition ? 
		       rule->condition->id : 
		       rule->condition_ref);
      action_ref = (rule->action ?
		    rule->action->id :
		    rule->action_ref);
    }
  
  store[0].dest = (void **) &id;
  store[1].dest = (void **) &condition_ref;
  store[2].dest = (void **) &action_ref;

  r = lxml_element_store_attr(element, store, op);

  if (op == LXML_Read)
    {
      if (r)
	{
	  A_LOGGING_ERROR("Rule syntax error. Ignored.");
	  goto err;
	}
      
      if (!id || !(condition_ref || action_ref))
	{
	  if (id)
	    A_LOGGING_ERROR("Rule '%s' does not have necessary attributes. "
			    "Ignored.", id);
	  else
	    A_LOGGING_ERROR("Rule is missing the id attribute");
	  goto err;
	}
      
      rule = a_Processing_rule_new();
      if (rule == NULL)
	goto err;
      
      rule->id = id;
      rule->condition_ref = condition_ref;
      rule->action_ref = action_ref;

      a_Processing_rule_add(rule);
    }

  return r;

 err:
  if (id)
    free(id);
  if (condition_ref)
    free(condition_ref);
  if (action_ref)
    free(action_ref);
  return -1;
}

int
a_Processing_update_rules_links(void)
{
  chashiter  *hiter;
  XePProcessingRule *rule;
  chashdatum datum;

  if (global_processing == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("Attempt to update rules links in "
			     "void processing list");
      return -1;
    }

  for (hiter = chash_begin(global_processing->rules); 
       hiter; 
       hiter = chash_next(global_processing->rules, hiter))
    {
      chash_value(hiter, &datum);
      rule = (XePProcessingRule *) datum.data;
      Processing_update_rule_links(rule);
    }
  return 0;
}

static int
Processing_update_rule_links(XePProcessingRule *rule)
{
  chashdatum key;
  chashdatum datum;
  A_LOGGING_DEBUG("Updating rule '%s' links", rule->id);

  rule->condition = NULL;
  rule->action    = NULL;
  if (rule->condition_ref)
    {
      key.data = rule->condition_ref;
      key.len  = strlen(key.data);
      if (global_processing->conditions && 
	  chash_get(global_processing->conditions, &key, &datum) == 0)
	{
	  rule->condition = (XePProcessingCondition *) datum.data;
	  free(rule->condition_ref);
	  rule->condition_ref = NULL;
	  A_LOGGING_DEBUG("Rule '%s' linked to condition '%s'",
			  rule->id, rule->condition->id);
	}
      else
	{
	  A_LOGGING_ERROR("Rule '%s' references unknown condition '%s'",
			  rule->id, rule->condition_ref);
	  return -1;
	}
      if (rule->action_ref)
	{
	  key.data = rule->action_ref;
	  key.len  = strlen(key.data);
	  if (global_processing->actions && 
	      (chash_get(global_processing->actions, &key, &datum) == 0))
	    {
	      rule->action = (XePProcessingAction *) datum.data;
	      free(rule->action_ref);
	      rule->action_ref = NULL;
	      A_LOGGING_DEBUG("Rule '%s' linked to action '%s'",
			      rule->id, rule->action->id);
	    }
	  else
	    {
	      A_LOGGING_ERROR("Rule '%s' references unknown action '%s'",
			      rule->id, rule->action_ref);
	      return -1;
	    }
	}
    }
  return 0;
}

static void *
Processing_get_from_ref(chash *hash, char *ref)
{
  chashdatum key;
  chashdatum datum;

  if (!hash || !ref)
    return NULL;

  key.data = ref;
  key.len  = strlen(key.data);

  if (chash_get(hash, &key, &datum))
    return NULL;

  return datum.data;
}

XePProcessingRule *
a_Processing_rule_get_from_ref(char *ref)
{
  return (XePProcessingRule *) Processing_get_from_ref(global_processing->rules, ref);
}

XePProcessingCondition *
a_Processing_condition_get_from_ref(char *ref)
{
  return (XePProcessingCondition *) Processing_get_from_ref(global_processing->conditions, ref);
}

int
a_Processing_ref_list_copy(clist *source, clist *target)
{
  clistiter *iter;
  XePProcessingRef *ref;
  XePProcessingRef *ref_copy;
  int r = 0;

  /* TODO: this is actually an append instead of a copy */
  for (iter = clist_begin(source); iter && r == 0; iter = iter->next)
    {
      ref = (XePProcessingRef *) iter->data;
      ref_copy = a_Processing_ref_new_copy(ref);
      if (ref_copy)
	{
	  if (clist_append(target, ref_copy))
	    {
	      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
	      a_Processing_ref_free(ref_copy);
	      r = -1;
	    }
	}
      else
	{
	  r = -1;
	}
    }
  return r;
}


static XePProcessingConditionTree *
Processing_condition_node_new(XePConditionNodeType node_type,
			      XePBooleanType boolean_type,
			      XePExpressionAttributeType attr_type, 
			      char *attr_header, 
			      XePExpressionTestType test_type,
			      XePExpressionValue *value)
{
  XePProcessingConditionTree *node;
  
  node = x_alloc(1, XePProcessingConditionTree);
  if (node == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }
  node->node_type = node_type;
  node->negation = 0;

  if (node_type == XEP_BOOLEAN)
    {
      node->type.boolean = boolean_type;
      node->op1.node = NULL;
      node->op2.node = NULL;
    }
  else
    {
      node->op1.type    = attr_type;
      node->attr_header = attr_header;
      node->type.test   = test_type;
      node->op2.value   = *value;
    }
  return node;
}

static void
Processing_condition_node_free(XePProcessingConditionTree *node)
{
  if (node == NULL)
    return;
  if (node->node_type == XEP_BOOLEAN)
    {
      Processing_condition_node_free(node->op1.node);
      Processing_condition_node_free(node->op2.node);
    }
  else
    {
      x_free(node->attr_header);
      if (node->op2.value.type == XEP_EXPRESSION_VALUE_STRING)
	{
	  x_free(node->op2.value.value.string);
	}
      if (node->op2.value.values)
	{
	  clistiter *iter;
	  for (iter = clist_begin(node->op2.value.values); iter; iter = iter->next)
	    {
	      x_free(iter->data);
	    }
	}
    }
  free(node);
}

#define PASS_SPACE(str, idx) \
  while (str[idx] == ' '  || str[idx] == '\t' ||	\
	 str[idx] == '\r' || str[idx] == '\n')		\
    { (idx)++; }

static XePProcessingConditionTree *
Processing_condition_parse_atom(const char *expr, size_t *index)
{
  XePProcessingConditionTree *node;
  size_t idx = *index;
  size_t start = *index;
  size_t val_start;
  int i;
  XePExpressionAttributeType attr_type;
  XePExpressionTestType test_type;
  XePExpressionValue value;
  char *attr_header = NULL;

  XePExpressionToken attributes [] = 
    {
      {"cc: ",      XEP_ATTRIBUTE_CC, XEP_EXPRESSION_VALUE_STRING},
      {"date: ",    XEP_ATTRIBUTE_DATE_H, XEP_EXPRESSION_VALUE_STRING},
      {"date ",     XEP_ATTRIBUTE_DATE_D, XEP_EXPRESSION_VALUE_DATE},
      {"folder ",   XEP_ATTRIBUTE_FOLDER, XEP_EXPRESSION_VALUE_STRING},
      {"from: ",    XEP_ATTRIBUTE_FROM,   XEP_EXPRESSION_VALUE_STRING},
      {"size ",     XEP_ATTRIBUTE_SIZE,   XEP_EXPRESSION_VALUE_SIZE},
      {"subject: ", XEP_ATTRIBUTE_SUBJECT, XEP_EXPRESSION_VALUE_STRING},
      {"uid ",      XEP_ATTRIBUTE_UID,     XEP_EXPRESSION_VALUE_UID},
      {"to: ",      XEP_ATTRIBUTE_TO,      XEP_EXPRESSION_VALUE_STRING},
      /* The following need message fetching */
      /* ".*: " user provided header */
      {"body ",     XEP_ATTRIBUTE_BODY,    XEP_EXPRESSION_VALUE_STRING},
      /* "mime_HEADER: " */
      /* "mime_size: " */
      /* The following can change throughout message life */
      {"age ",      XEP_ATTRIBUTE_AGE,     XEP_EXPRESSION_VALUE_AGE},
      {"score ",    XEP_ATTRIBUTE_SCORE,   XEP_EXPRESSION_VALUE_NUMBER},
      {"flags ",    XEP_ATTRIBUTE_FLAGS,   XEP_EXPRESSION_VALUE_FLAGS},
      /* The following are beyond control */
      {"shell ",    XEP_ATTRIBUTE_SHELL,   XEP_EXPRESSION_VALUE_SHELL},
      {"python ",   XEP_ATTRIBUTE_PYTHON,  XEP_EXPRESSION_VALUE_PYTHON},
      {"filter ",   XEP_ATTRIBUTE_FILTER,  XEP_EXPRESSION_VALUE_FILTER},
      {NULL,        XEP_ATTRIBUTE_HEADER,  XEP_EXPRESSION_VALUE_STRING}
    };
  int attr_count = sizeof(attributes)/sizeof(attributes[0]);
  XePExpressionTest string_tests [] = 
    {
      /* Third column ignored */
      {"is ",     XEP_EXPRESSION_TEST_IS,                  },
      {"IS ",     XEP_EXPRESSION_TEST_IS_CASESENSITIVE,    },
      {"has ",    XEP_EXPRESSION_TEST_HAS,                 },
      {"HAS ",    XEP_EXPRESSION_TEST_HAS_CASESENSITIVE,   },
      {"regexp ", XEP_EXPRESSION_TEST_REGEXP,              },
      {"REGEXP ", XEP_EXPRESSION_TEST_REGEXP_CASESENSITIVE,}
    };
  int string_test_count = sizeof(string_tests)/sizeof(string_tests[0]);
  XePExpressionTest order_tests[] = 
    {
      /* Third column ignored */
      {"< ",  XEP_EXPRESSION_TEST_LT},
      {"<= ", XEP_EXPRESSION_TEST_LE},
      {"> ",  XEP_EXPRESSION_TEST_GT},
      {">= ", XEP_EXPRESSION_TEST_GE},
      {"= ",  XEP_EXPRESSION_TEST_EQ},
      {"<> ", XEP_EXPRESSION_TEST_NE}
    };
  int order_test_count = sizeof(order_tests)/sizeof(order_tests[0]);
  XePExpressionTest *tests;
  int test_count;

  /* Initialize value data that would need freeing */
  value.value.string = NULL;
  value.values = NULL;

  PASS_SPACE(expr, idx);
      
  /* Expression is of the form ATTRIBUTE OPERATOR VALUE 
   * An alternative form is ATTRIBUTE VALUE1 VALUE2... depending on ATTRIBUTE
   */

  /* ATTRIBUTE */
  attr_type = XEP_ATTRIBUTE_NONE;
  for (i = 0; i < attr_count; i++)
    {
      if (attributes[i].type == XEP_ATTRIBUTE_HEADER)
	{
	  /* Search for user defined headers with the ".*: " syntax */
	  while (expr[idx] && expr[idx] != ' ' && expr[idx] != ':')
	    {
	      idx++;
	    }
	  
	  if (expr[idx] == ':')
	    {
	      attr_type = attributes[i].type;
	      value.type  = attributes[i].value_type;
	      idx++;
	      attr_header = x_alloc((idx - start + 1) + 1, char);
	      memcpy(attr_header, expr + start, (idx - start + 1));
	      attr_header[idx - start + 1] = 0x00;
	      break;
	    }
	}
      else
	{
	  /* Search for known attributes tokens */
	  if (strncmp(expr + idx, attributes[i].token, strlen(attributes[i].token)) == 0)
	    {
	      attr_type = attributes[i].type;
	      value.type  = attributes[i].value_type;
	      idx += strlen(attributes[i].token);
	      break;
	    }
	}
    }

  if (attr_type == XEP_ATTRIBUTE_NONE)
    {
      A_LOGGING_ERROR(_("Syntax error: unknown attribute <%s>"), expr + start);
      x_free(attr_header);
      return NULL;
    }

  /* OPERATOR */
  /* Depending on attribute value, we check for an operator and value */

  switch (value.type)
    {
    case XEP_EXPRESSION_VALUE_STRING:
      /* Operands are: is has regexp IS HAS REGEXP */
      tests = string_tests;
      test_count =string_test_count;
      break;
    case XEP_EXPRESSION_VALUE_SIZE:
    case XEP_EXPRESSION_VALUE_DATE:
    case XEP_EXPRESSION_VALUE_NUMBER:
    case XEP_EXPRESSION_VALUE_UID:
    case XEP_EXPRESSION_VALUE_AGE:
      tests = order_tests;
      test_count = order_test_count;
      break;
    case XEP_EXPRESSION_VALUE_SHELL:
    case XEP_EXPRESSION_VALUE_PYTHON:
    case XEP_EXPRESSION_VALUE_FILTER:
      tests = NULL;
      test_count = 0;
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("expression value type unknown in %s", expr + start);
      x_free(attr_header);
      return NULL;
    }

  /* Parse operator if required */
  PASS_SPACE(expr, idx);

  test_type = XEP_EXPRESSION_TEST_NONE;

  for (i = 0; i < test_count; i++)
    {
      if (strncmp(expr + idx, tests[i].token, strlen(tests[i].token)) == 0)
	{
	  test_type = tests[i].type;
	  idx += strlen(tests[i].token);
	  break;	  
	}
    }

  if (test_count > 0 && test_type == XEP_EXPRESSION_TEST_NONE)
    {
      A_LOGGING_ERROR(_("Syntax error: operator missing in expression <%s>"), expr + start);
      x_free(attr_header);
      return NULL;
    }

  /* VALUE (OPERAND) */
  PASS_SPACE(expr, idx);

  switch (value.type)
    {
    case XEP_EXPRESSION_VALUE_STRING:
      if (expr[idx] != '"')
	{
	  A_LOGGING_ERROR(_("Syntax error: string value missing in expression <%s>"), expr + start);
	  x_free(attr_header);
	  return NULL;
	}
      idx++;
      val_start = idx;
      while (expr[idx] && expr[idx] != '"')
	{
	  if (expr[idx] == '\\')
	    {
	      /* Skip escaped character */
	      idx++;
	      if (expr[idx] == 0x00)
		{
		  break;
		}
	    }
	  idx++;
	}
      if (expr[idx] == 0x00)
	{
	  A_LOGGING_ERROR(_("Syntax error: unbounded string found in expression <%s>"), expr + start);
	  x_free(attr_header);
	  return NULL;
	}
      value.value.string = x_alloc(idx - val_start + 1, char);
      memcpy(value.value.string, expr + val_start, idx - val_start);
      value.value.string[idx - val_start] = 0x00;
      idx++;
      break;
    case XEP_EXPRESSION_VALUE_UID:
      if ((expr[idx] < '0' || expr[idx] > '9'))
	{
	  A_LOGGING_ERROR(_("Syntax error: unsigned number for uid value expected in expression <%s>"), expr + start);
	  x_free(attr_header);
	  return NULL;
	}
      {
	char *endptr;
	value.value.uid = strtoul(expr + idx, &endptr, 0);
	/* we already checked for number validity, so no need
	 * to check endptr == expr + idx
	 */
	idx = endptr - expr;
      }
      break;
    case XEP_EXPRESSION_VALUE_SIZE:
    case XEP_EXPRESSION_VALUE_NUMBER:
      if (expr[idx] != '+' && expr[idx] != '-' && (expr[idx] < '0' || expr[idx] > '9'))
	{
	  A_LOGGING_ERROR(_("Syntax error: number value expected in expression <%s>"), expr + start);
	  x_free(attr_header);
	  return NULL;
	}
      {
	char *endptr;
	value.value.number = strtol(expr + idx, &endptr, 0);
	/* we already checked for number validity, so no need
	 * to check endptr == expr + idx
	 */
	idx = endptr - expr;
      }
      if (value.type == XEP_EXPRESSION_VALUE_SIZE)
	{
	  if (expr[idx] == 'k')
	    {
	      value.value.number *= 1024;
	      idx++;
	    }
	  else if (expr[idx] == 'M')
	    {
	      value.value.number *= 1024*1024;
	      idx++;
	    }
	  else if (expr[idx] == 'G')
	    {
	      value.value.number *= 1024*1024*1024;
	      idx++;
	    }
	  if (expr[idx] == 'b')
	    {
	      idx++;
	    }
	}
      if (expr[idx] != ' ' && expr[idx] != ')' && expr[idx])
	{
	  A_LOGGING_ERROR(_("Syntax error: garbage after number value in <%s>"), expr + start);
	  x_free(attr_header);
	  return NULL;
	}
      break;
    default:
      A_LOGGING_ERROR("not implemented yet...");
    }
  
  node = Processing_condition_node_new(XEP_EXPRESSION, XEP_BOOLEAN_NONE, 
				       attr_type, attr_header, test_type, &value);

  PASS_SPACE(expr, idx);

  *index = idx;
  return node;
}

static XePProcessingConditionTree *
Processing_condition_parse_expression(const char *expr, size_t *index, int *depth)
{
  XePProcessingConditionTree *node;
  int negation = 0;
  size_t idx = *index;
  const char *XEP_NOT = "not";
  const size_t XEP_NOT_LEN = strlen(XEP_NOT);

  PASS_SPACE(expr, idx);

  if (strlen(expr + idx) >= 4 && strncmp(expr + idx, XEP_NOT, XEP_NOT_LEN) == 0 &&
      (expr[idx + 3] == ' ' || expr[idx + 3] == '('))
    {
      negation = 1;
      idx += XEP_NOT_LEN;
      PASS_SPACE(expr, idx);
    }

  if (expr[idx] == '(')
    {
      idx++;
      PASS_SPACE(expr, idx);
      *index = idx;
      printf("Depth at %d\n", *depth);
      (*depth)++;
      printf("Depth inc to %d\n", *depth);
      node = Processing_condition_compile(expr, index, depth);
      if (node && negation)
	node->negation = !node->negation;
    }
  else
    {
      *index = idx;
      node = Processing_condition_parse_atom(expr, index);
      if (node)
	node->negation = negation;
    }
  return node;
}

static XePProcessingConditionTree *
Processing_condition_compile(const char *expr, size_t *index, int *depth)
{
  XePProcessingConditionTree *node;
  XePProcessingConditionTree *next_node;
  XePProcessingConditionTree *bool_node;
  XePProcessingConditionTree *next_bool_node;
  int pri;
  int next_pri;
  size_t start = *index;
  size_t idx;
  XePBooleanType boolean;
  XePBooleanType next_boolean;

  /* ...X(i) BOOLi X(i+1) BOOL(i+1)... */
  /* Parse X(i) */
  node = Processing_condition_parse_expression(expr, index, depth);
  if (node == NULL)
    return NULL;

  idx = *index;

  /* End of string? Then X(i) was the last expression */
  if (expr[idx] == 0x00)
    {
      return node;
    }
  
  /* End of parenthesized expression? Then return node X(i) */
  if (expr[idx] == ')')
    {
      idx++;
      PASS_SPACE(expr, idx);
      printf("Depth at %d\n", *depth);
      (*depth)--;
      printf("Depth dec to %d\n", *depth);
      *index = idx;
      if (*depth < 0)
	{
	  A_LOGGING_ERROR(_("Syntax error: unmatched closing parenthesis at <%s>"), expr + idx);
	  Processing_condition_node_free(node);
	  return NULL;
	}
      return node;
    }

  /* BOOL(i): boolean operator */
  boolean = (expr[idx] == '&' ?
	     XEP_BOOLEAN_AND :
	     expr[idx] == '|' ?
	     XEP_BOOLEAN_OR :
	     expr[idx] == '^' ?
	     XEP_BOOLEAN_XOR :
	     XEP_BOOLEAN_NONE);

  if (boolean == XEP_BOOLEAN_NONE)
    {
      A_LOGGING_ERROR(_("Syntax error: unknown boolean <%c> operator in <%s>"), boolean, expr + start);
      *index = idx;
      Processing_condition_node_free(node);
      return NULL;
    }

  idx++;

  PASS_SPACE(expr, idx);
  
  /* We test only for AND, because the rest of supported booleans are OR and XOR which are
   * of lower priority
   */
  pri = (boolean == XEP_BOOLEAN_AND ? 1 : 0);

  /* Create node for BOOL(i):
   *     BOOL(i)
   *    / 
   *  X(i)
   */
  bool_node = Processing_condition_node_new(XEP_BOOLEAN, boolean, 
					    XEP_ATTRIBUTE_NONE, NULL, XEP_EXPRESSION_TEST_NONE, NULL);
  bool_node->op1.node = node;

  while (1)
    {
      /* X(i+1) */
      *index = idx;
      next_node = Processing_condition_parse_expression(expr, index, depth);
      idx = *index;
      if (next_node == NULL)
	{
	  Processing_condition_node_free(next_node);
	  Processing_condition_node_free(bool_node);
	  return NULL;
	}

      /* If X(i+1) was last expression, then finish BOOL(i) and return it:
       *     BOOL(i)
       *    /      \
       *  X(i)     X(i+1)
       */
      if (expr[idx] == 0x00)
	{
  bool_node->op2.node = next_node;
	  return bool_node;
	}

      /* If X(i+1) was last expression of a parenthized succession, finish BOOL(i) (as above) and return it:
       *     BOOL(i)
       *    /      \
       *  X(i)     X(i+1)
       */
      if (expr[idx] == ')')
	{
	  (*depth)--;
	  printf("*Depth dec to %d\n", *depth);
	  if (*depth < 0)
	    {
	      A_LOGGING_ERROR(_("Syntax error: unmatched closing parenthesis at <%s>"), expr + idx);
	      Processing_condition_node_free(next_node);
	      Processing_condition_node_free(bool_node);
	      return NULL;
	    }
	  idx++;
	  PASS_SPACE(expr, idx);
	  *index = idx;
	  bool_node->op2.node = next_node;
	  return bool_node;
	}
       
       /* Check BOOL(i+1) */
       next_boolean = (expr[idx] == '&' ?
		       XEP_BOOLEAN_AND :
		       expr[idx] == '|' ?
		       XEP_BOOLEAN_OR :
		       expr[idx] == '^' ?
		       XEP_BOOLEAN_XOR :
		       XEP_BOOLEAN_NONE);
       
       if (next_boolean == XEP_BOOLEAN_NONE)
	 {
	   A_LOGGING_ERROR(_("Syntax error: unknown boolean <%c> operator in <%s>"), expr[idx], expr + start);
	   *index = idx;
	   Processing_condition_node_free(next_node);
	   Processing_condition_node_free(bool_node);
	   return NULL;
	 }
       
       next_pri = (next_boolean == XEP_BOOLEAN_AND ? 1 : 0);
       
       idx++;
       PASS_SPACE(expr, idx);
       
       next_bool_node = Processing_condition_node_new(XEP_BOOLEAN, next_boolean, 
						      XEP_ATTRIBUTE_NONE, NULL, XEP_EXPRESSION_TEST_NONE, NULL);
       /* BOOL(i+1) has lower or equal priority than previous BOOL(i):
	*
	*          BOOL(i+1)
	*         /
	*      BOOL(i)
	*     /      \
	*   X(i)    X(i+1)
	*
	*  Then, i = i + 1;
	*/
       if (next_pri <= pri)
	 {
	   bool_node->op2.node = next_node;
	   next_bool_node->op1.node = bool_node;
	   bool_node = next_bool_node;
	   pri = next_pri;
	   continue;
	 }
       /* next_pri > pri */
       /* BOOL(i+1) has a greater priority than previous BOOL(i), eval at (i+2) and return BOOL(i)
	*
	*    BOOL(i)
	*   /      \
	* X(i)  BOOL(i+1)
	*      /        \
	*    X(i+1)  NODE(i+2)
	*
	*/
       next_bool_node->op1.node = next_node;
       bool_node->op2.node = next_bool_node;
       *index = idx;
       next_bool_node->op2.node = Processing_condition_parse_expression(expr, index, depth);
       
       if (next_bool_node->op2.node == NULL)
	 {
	   Processing_condition_node_free(bool_node);
	   bool_node = NULL;
	 }
       return bool_node;
    }
}

int
a_Processing_rules_compile_all(void)
{
  chashdatum datum;
  chashiter  *hiter;
  chash *rules;
  int r = 0;
  XePProcessingRule *rule;

  rules = global_processing->rules;

  for (hiter = chash_begin(rules); hiter; hiter = chash_next(rules, hiter))
    {
      chash_value(hiter, &datum);
      rule = (XePProcessingRule *) datum.data;

      if (a_Processing_condition_compile(rule->condition))
	{
	  A_LOGGING_ERROR(_("Processing rule <%s> could not be compiled. Rule Disabled"), rule->id);
	  r = 1;
	}
    }
  return r;
}

int
a_Processing_condition_compile(XePProcessingCondition *condition)
{
  int depth = 0;
  size_t index = 0;

  if (condition->expression == NULL)
    return -1;
  
  if (condition->tree)
    return 0;

  A_LOGGING_DEBUG("Compiling condition <%s>...", condition->expression);

  condition->tree = Processing_condition_compile(condition->expression, &index, &depth);

  A_LOGGING_DEBUG("... compiled: %d", (condition->tree != NULL));
  return (condition->tree == NULL);
}

/* We need the macro because we have many types of numbers:
 * size_t, long int and uint32_t (see FIXME below)
 */
#define RETURN_ORDER_COMPARISON(test_type, op1, op2)	\
  switch(test_type) \
    { \
    case XEP_EXPRESSION_TEST_LT: \
      return op1 < op2; \
    case XEP_EXPRESSION_TEST_LE: \
      return op1 <= op2; \
    case XEP_EXPRESSION_TEST_GT: \
      return op1 > op2; \
    case XEP_EXPRESSION_TEST_GE: \
      return op1 >= op2; \
    case XEP_EXPRESSION_TEST_EQ: \
      return op1 == op2; \
    case XEP_EXPRESSION_TEST_NE: \
      return op1 != op2; \
    }

int
a_Processing_message_match(XePProcessingConditionTree *condition,
			   XePMessage *message)
{
  char *text, *c_text;
  size_t size, c_size;
  struct tm *c_date;
  uint32_t c_flags_on, c_flags_off;
  long int number, c_number;
  uint32_t uid, c_uid;
  mailmessage *msg;
  XePExpressionTestType test_type;
  int r;

  if (condition->node_type == XEP_BOOLEAN)
    {
      int r1, r;
	  
      /* TODO: We should first evaluate the node that does not require 
       * message fetching, and not necessarily node1
       */
      r1 = a_Processing_message_match(condition->op1.node, message);

      if ((!r1 && condition->type.boolean == XEP_BOOLEAN_AND) ||
	  ( r1 && condition->type.boolean == XEP_BOOLEAN_OR))
	{
	  return (condition->negation ? !r1 : r1);
	}

      r = a_Processing_message_match(condition->op2.node, message);
      
      if (condition->type.boolean != XEP_BOOLEAN_XOR)
	return (condition->negation ? !r : r);

      /* XEP_BOOLEAN_XOR */
      r = !(r == r1);
      return (condition->negation ? !r : r);
    }

  /* node_type = XEP_EXPRESSION. Check if message matches */
  text = NULL;
  test_type = condition->type.test;
  msg = message->msg;

  switch (condition->op1.type)
    {
    case XEP_ATTRIBUTE_SUBJECT:
      a_Message_message_get_subject(&msg->single_fields, "UTF-8", &text);
      break;
    case XEP_ATTRIBUTE_FROM:
      {
	char *d = NULL, *e = NULL;
	size_t ld = 0, le = 0;
	a_Message_message_get_from(&msg->single_fields, "UTF-8", &d, &e);
	if (d)
	  ld = strlen(d);
	if (e)
	  le = strlen(e);
	text = malloc(ld + le + 1);
	if (d)
	  {
	    memcpy(text, d, ld);
	    free(d);
	  }
	if (e)
	  {
	    memcpy(text + ld, e, le);
	    free(e);
	  }
	text[ld + le] = 0x00;
      }
      break;

    case XEP_ATTRIBUTE_SIZE:
      size = message->msg->size;
      break;
    case XEP_ATTRIBUTE_UID:
      uid = message->msg->index;
      break;
    default:
      A_LOGGING_ERROR("Search not implemented");
      return 0;
    }
  switch (condition->op2.value.type)
    {
    case XEP_EXPRESSION_VALUE_SIZE:
      c_size = condition->op2.value.value.size;
      RETURN_ORDER_COMPARISON(test_type, size, c_size);
    case XEP_EXPRESSION_VALUE_NUMBER:
      c_number = condition->op2.value.value.number;
      RETURN_ORDER_COMPARISON(test_type, number, c_number);
      break;
    case XEP_EXPRESSION_VALUE_UID:
      /* FIXME: add a uint32_t field for uid in value */
      c_uid = (uint32_t) condition->op2.value.value.number;
      RETURN_ORDER_COMPARISON(test_type, uid, c_uid);
    case XEP_EXPRESSION_VALUE_STRING:
      c_text = condition->op2.value.value.string;
      /* FIXME: string comparison under UTF-8 is tricky. */
      r = 0;
      if (text)
	{
	  r = (strstr(text, c_text) != NULL);
	  free(text);
	}
      return r;
      break;
    case XEP_EXPRESSION_VALUE_DATE:
      c_date = &condition->op2.value.value.date;
      A_LOGGING_ERROR("Date comparison not implemented yet");
      return 0;
      break;
    case XEP_EXPRESSION_VALUE_FLAGS:
      c_flags_on = condition->op2.value.value.flags.flags_on;
      c_flags_off = condition->op2.value.value.flags.flags_off;
      return (((msg->flags->flags & c_flags_on) == c_flags_on) &&
	      ((msg->flags->flags & c_flags_off) == 0));
      break;
    default:
      /* TODO */
      A_LOGGING_ERROR("Expression value type in comparison not implemented yet");
    }
  return 0;
}

int
a_Processing_message_action(XePProcessingAction *action,
			    XePMessage *message)
{
  printf("Message %d matches!\n", message->msg->index);
  return 0;
}
